""" A python module for accessing BioMart MartService

Example::
    >>> connection = BioMartConnection("http://www.biomart.org/biomart/martservice")
    >>> reg = BioMartRegistry(connection)
    >>> for mart in reg.marts():
    ...    print mart.name
    ...
    ensembl...
    >>> dataset = BioMartDataset(mart="ensembl", internalName="hsapiens_gene_ensembl", virtualSchema="default", connection=connection)
    >>> for attr in dataset.attributes()[:10]:
    ...    print attr.name
    ...
    Ensembl Gene ID...
    >>> data = dataset.get_data(attributes=["ensembl_gene_id", "ensembl_peptide_id"], filters=[("chromosome_name", "1")])
    
    >>> query = BioMartQuery(reg.connection, virtualSchema="default")
    >>> query.set_dataset("hsapiens_gene_ensembl")
    >>> query.add_attribute("ensembl_gene_id")
    >>> query.add_attribute("ensembl_peptide_id")
    >>> query.add_filter("chromosome_name", "1")
    >>> count = query.get_count()
"""

import sys, os
import urllib2
import traceback
import posixpath
import xml.dom.minidom as minidom
import shelve
import itertools

import warnings

from functools import partial

from Orange.orng import orngEnviron

class BioMartError(Exception):
    pass

class BioMartServerError(BioMartError):
    pass

class DatabaseNameConflictError(BioMartError):
    pass

class BioMartQueryError(BioMartError):
    pass

def _init__slots_from_line(self, line):
    for attr, val in zip(self.__slots__, line.split("\t")):
        setattr(self, attr, val)
        
class Attribute(object):
    """ A class representing an attribute in a BioMart dataset (returned by BioMartDataset.attributes())
    Attributes:
        - *internalName*    - Internal attribute name
        - *name*    - Human readable name of the attribute
        - *description*    - Human readable description of the attribute
    """
    __slots__ = ["internalName", "name", "description", "_4", "format", "table", "_7"]
    __init__ = _init__slots_from_line
    
    def __repr__(self):
        return "Attribute('%s')" % "\t".join(getattr(self, name, "") for name in self.__slots__)
     
            
class Filter(object):
    """ A class representing a filter for a BioMart dataset (returned by BioMartDataset.filters())
    Attributes:
        - *internalName* - Internal filter name
        - *name*    - Filter name
        - *values*    - Lists possible filter values
        - *description*    - Filter description
    """
    __slots__ =  ["internalName", "name", "values", "description", "type", "qualifiers", "_6", "attribute_table"]
    __init__ = _init__slots_from_line
    
    def __repr__(self):
        return "Filter('%s')" % "\t".join(getattr(self, name, "") for name in self.__slots__)


class XMLNode(object):
    def __init__(self, tag, attributes, children=[], data=""):
        self.tag = tag
        self.attributes = attributes
        self.children = children if children else []
        self.data = data if data else ""
        
    @classmethod    
    def fromDOMNode(cls, node):
        return XMLNode(node.tagName, dict(node.attributes.items()))
        
    def _match(self, tag=None, **kwargs):
        match = True
        if tag is not None and self.tag != tag:
            match = False
        
        return match and all(self.attributes.get(key, None) == value for key, value in kwargs.iteritems())
                    
    def elements(self, tag=None, **kwargs):
        if self._match(tag, **kwargs):
            yield self
            
        for child in self.children:
            for el in child.elements(tag, **kwargs):
                yield el
                
    def elements_top(self, tag=None, **kwargs):
        """ Return only the top elements (do not return matching children
        of an already matching element)."
        """
        if self._match(tag, **kwargs):
            yield self
        else:
            for child in self.children:
                for el in child.elements_top(tag, **kwargs):
                    yield el
                    
    def subelements(self, *args, **kwargs):
        return itertools.chain(*[c.elements(*args, **kwargs) for c in self.children])
    
    def subelements_top(self, *args, **kwargs):
        return itertools.chain(*[c.elements_top(*args, **kwargs) for c in self.children])
                
    def __iter__(self):
        """ Iterate over all sub elements.
        """
        return itertools.chain(*[iter(c) for c in self.children])
                
    def __repr__(self):
        return 'XMLNode("%s", %s)' % (self.tag, repr(self.attributes))
                
def parseXML(stream, parser=None):
    import xml.dom.pulldom as pulldom
    if isinstance(stream, basestring):
        events = pulldom.parseString(stream, parser)
    else:
        events = pulldom.parse(stream, parser)
    
    document = None
    chain = []
    for event, node in events:
        if event == "START_DOCUMENT":
            chain.append(XMLNode("DOCUMENT", {}))
        
        elif event == "START_ELEMENT":
            
            node = XMLNode.fromDOMNode(node)
            if chain:
                chain[-1].children.append(node)
            chain.append(node)
            
        elif event == "END_ELEMENT":
            chain.pop(-1)
            
        elif event == "CHARACTERS":
            chain[-1].data += node.data
            
        elif event == "END_DOCUMENT":
            document = chain.pop(-1)
    return document or chain[0]
    

def de_tab(text, sep="\t"):
    return [line.split(sep) for line in text.split("\n") if line.strip()]


def cached(func):
    from functools import wraps
    @wraps(func)
    def f(self):
        if getattr(self, "_cache_" + func.__name__, None) is None:
            setattr(self, "_cache_" + func.__name__, func(self))
        return getattr(self, "_cache_" + func.__name__)
    return f

    
def drop_iter(generator):
    """ Drop all elements from the iterator that raise an exception  
    """
    def _iter(generator):
        while True:
            try:
                yield generator.next()
            except StopIteration:
                raise StopIteration
            except Exception, ex:
                warnings.warn("An error occured during iteration:\n%s" % str(ex), UserWarning, stacklevel=2)
    return list(_iter(generator))

    
DEFAULT_ADDRESS = "http://www.biomart.org/biomart/martservice"

try:
    DEFAULT_DATA_CACHE = shelve.open(os.path.join(orngEnviron.bufferDir, "BioMartDataCache.pck"))
except Exception, ex:
    warnings.warn("Could not open BioMart data cache! %s" % str(ex))
    DEFAULT_DATA_CACHE = {}
    
try:
    DEFAULT_META_CACHE = shelve.open(os.path.join(orngEnviron.bufferDir, "BioMartMetaCache.pck"))
except Exception, ex:
    warnings.warn("Could not open BioMart meta cache! %s" % str(ex))
    DEFAULT_META_CACHE = {}


def checkBioMartServerError(response):
    if response.strip().startswith("Mart name conflict"):
        raise DatabaseNameConflictError(response)
    elif response.strip().startswith("Problem retrieving datasets"):
        raise BioMartError(response)
    elif response.startswith("non-BioMart die():"):
        raise BioMartServerError(response)


class BioMartConnection(object):
    """ A connection to a BioMart martservice server.
    
    Example::
        >>> connection = BioMartConnection("http://www.biomart.org/biomart/martservice")
        >>> response = connection.registry()
        >>> response = connection.datasets(mart="ensembl")
    """
    FOLLOW_REDIRECTS = False
    
    
    def __init__(self, address=None, cache=None, metaCache=None, dataCache=None):
        metaCache = cache if metaCache is None else metaCache
        dataCache = cache if dataCache is None else dataCache
        
        self.address = address if address is not None else DEFAULT_ADDRESS
#        self.cache = cache if cache is not None else DEFAULT_CACHE
        self.metaCache = metaCache if metaCache is not None else DEFAULT_META_CACHE
        self.dataCache = dataCache if dataCache is not None else DEFAULT_DATA_CACHE
        self.errorCache = {}
    
    def _get_cache_for_args(self, kwargs):
        if "type" in kwargs:
            return self.metaCache
        elif "query" in kwargs:
            return self.dataCache
        
    def request_url(self, **kwargs):
        order = ["type", "dataset", "mart", "virtualSchema", "query"]
        items = sorted(kwargs.iteritems(), key=lambda item: order.index(item[0]) if item[0] in order else 10)
        url = self.address + "?" + "&".join("%s=%s" % item for item in items if item[0] != "POST")
        return url.replace(" ", "%20")
    
    def request(self, **kwargs):
        url = self.request_url(**kwargs)
        
        cache = self._get_cache_for_args(kwargs)
        
        if url in self.errorCache:
            raise self.errorCache[url]
        
        if str(url) not in cache:
            try:
                response = urllib2.urlopen(url).read()
                checkBioMartServerError(response)
            except Exception, ex:
                self.errorCache[url] = ex
                raise ex
            
            cache[str(url)] = response
            if hasattr(cache, "sync"):
                cache.sync()
                
        from StringIO import StringIO
        return StringIO(cache[str(url)])
    
    def registry(self, **kwargs):
        return self.request(type="registry")
    
    def datasets(self, mart="ensembl", **kwargs):
        return self.request(type="datasets", mart=mart, **kwargs)
    
    def attributes(self, dataset="oanatinus_gene_ensembl", **kwargs):
        return self.request(type="attributes", dataset=dataset, **kwargs)
    
    def filters(self, dataset="oanatinus_gene_ensembl", **kwargs):
        return self.request(type="filters", dataset=dataset, **kwargs)
    
    def configuration(self, dataset="oanatinus_gene_ensembl", **kwargs):
        return self.request(type="configuration", dataset=dataset, **kwargs)
    
    def clearCache(self):
        self.dataCache.clear()
        self.metaCache.clear()
        self.errorCache.clear()    
        
        
class BioMartRegistry(object):
    """ A class representing a BioMart registry
    Arguments:
        - stream: A file like object with xml registry or a BioMartConnection instance
    
    Example::
    >>> registry = BioMartRegistry(connection)
    >>> for schema in registry.virtual_schemas():
    ...    print schema.name
    ...
    default
    """
    def __init__(self, stream):
        self.connection = stream if isinstance(stream, BioMartConnection) else None
        if self.connection:
            if hasattr(self.connection, "_registry"):
                self.__dict__ = self.connection._registry.__dict__
                return
            else:
                stream = self.connection.registry()
                self.connection._registry = self
        else:
            stream = open(stream, "rb") if isinstance(stream, basestring) else stream
        self.registry = self.parse(stream)
        
    
    @cached
    def virtual_schemas(self):
        """ Return a list of BioMartVirtualSchema instances representing each schema
        """
        schemas = [schema.attributes.get(name, "default") for name in self.registry.elements("virtualSchema")]
        if not schemas:
            schemas = [BioMartVirtualSchema(self.registry, name="default", connection=self.connection)]
        else:
            schemas = [BiomartVirtualSchema(schema, name=schema.attributes.get("name", "default"),
                            connection=self.connection) for schema in self.registry.elements("virtualSchema")]
        return schemas
    
    def virtual_schema(self, name):
        """ Return a named virtual schema  
        """
        for schema in self.virtual_schemas():
            if schema.name == name:
                return schema
        raise ValueError("Unknown schema name '%s'" % name)
        
    @cached
    def marts(self):
        """ Return a list off all 'mart' instances (BioMartDatabase instances) regardless of their virtual schemas  
        """
        return reduce(list.__add__, drop_iter(schema.marts() for schema in self.virtual_schemas()), [])
    
    def mart(self, name):
        """ Return a named mart.
        """
        for mart in self.marts():
            if mart.name == name:
                return mart
        raise ValueError("Unknown mart name '%s'" % name)
    
    def databases(self):
        """ Same as marts()
        """
        return self.marts()
    
    def datasets(self):
        """ Return a list of all datasets (BioMartDataset instances) from all marts regardless of their virtual schemas
        """
        return reduce(list.__add__, drop_iter(mart.datasets() for mart in self.marts()), [])
    
    def dataset(self, internalName, virtualSchema=None):
        """ Return a BioMartDataset instance that matches the internalName
        """
        if virtualSchema is not None:
            schema = self.virtual_schema(virtualSchema)
            return schema.dataset(internalName)
        else:
            for d in self.datasets():
                if d.internalName == internalName:
                    return d
            raise ValueError("Unknown dataset name: '%s'" % internalName)
            
    
    def query(self, **kwargs):
        """ Return an initialized BioMartQuery object with registry set to self.
        Pass additional arguments to BioMartQuery.__init__ with keyword arguments
        """
        return BioMartQuery(self, **kwargs)
    
    def links_between(self, exporting, importing, virtualSchema="default"):
        """ Return all links between exporting and importing datasets in the virtualSchema
        """
        schema = [schema for schema in self.virtual_schemas() if schema.name == virtualSchema][0]
        return schema.links_between(exporting, importing)
    
    def get_path(self, exporting, importing):
        raise NotImplementedError
    
    def __iter__(self):
        return iter(self.marts())
    
    def _find_mart(self, name):
        try:
            return [mart for mart in self.marts() if name in [getattr(mart, "name", None),
                                    getattr(mart, "internalName", None)]][0]
        except IndexError, ex:
            raise ValueError(name)
        
    def __getitem__(self, name):
        return self._find_mart(name)
    
    def search(self, string, relevance=False):
        results = []
        for mart in self.marts():
            results.extend([(rel, mart, dataset) for rel, dataset in  mart.search(string, True)])
        results = sorted(results, reverse=True)
        if not relevance:
            results = [(mart, dataset) for _, mart, dataset in results]
        return results
    
    @classmethod
    def parse(cls, stream, parser=None):
        """ Parse the registry file like object and return a DOM like description (XMLNode instance)
        """
        xml = stream.read()
        doc = parseXML(xml, parser)
        return doc
         
parseRegistry = BioMartRegistry.parse

class BioMartVirtualSchema(object):
    """ A class representation of a virtual schema.
    """
    def __init__(self, locations=None, name="default", connection=None):
        self.locations = locations
        self.name = name
        self.connection = connection
        
    @cached
    def marts(self):
        """ Return a list off all 'mart' instances (BioMartDatabase instances) in this schema
        """
        return drop_iter(BioMartDatabase(connection=self.connection, **dict((str(key), val) \
                for key, val in loc.attributes.items())) for loc in self.locations.elements("MartURLLocation"))
    
    def databases(self):
        """ Same as marts()
        """
        return self.marts()
    
    @cached
    def datasets(self):
        """ Return a list of all datasets (BioMartDataset instances) from all marts in this schema
        """
        return reduce(list.__add__, drop_iter(mart.datasets() for mart in self.marts()), [])
    
    
    def dataset(self, internalName):
        """ Return a dataset with matching internalName
        """
        for dataset in self.datasets():
            if dataset.internalName == internalName:
                return dataset
        raise ValueError("Unknown data set name '%s'" % internalName)
    
    
    def links_between(self, exporting, importing):
        """ Return a list of link names from exporting dataset to importing dataset 
        """
        exporting = self[exporting]
        importing = self[importing]
        exporting = exporting.configuration().exportables()
        importing = importing.configuration().importables()
        exporting = set([(ex.linkName, getattr(ex, "linkVersion", "")) for ex in exporting])
        importing = set([(im.linkName, getattr(ex, "linkVersion", "")) for im in importing])
        links = exporting.intersection(importing)
        return [link[0] for link in links]
    
    def links(self):
        """ Return a list of (linkName, linkVersion) tuples defined by datasets in the schema
        """
        pass
    
    def query(self, **kwargs):
        """ Return an initialized BioMartQuery object with registry and virtualSchema set to self.
        Pass additional arguments to BioMartQuery.__init__ with keyword arguments
        """
        return BioMartQuery(self, virtualSchema=self.name, **kwargs)

    def __iter__(self):
        return iter(self.marts())
    
    def __getitem__(self, key):
        try:
            return self.dataset(key)
        except ValueError:
            raise KeyError(key)
        
    
class BioMartDatabase(object):
    """ An object representing a BioMart 'mart' instance.
    Arguments::
        - *name*   - Name of the mart instance ('ensembl' by default)
        - *virtualSchema*    - Name of the virtualSchema this dataset belongs to ("default" by default)
        - *connection*    - An optional BioMartConnection instance
        
    """
    host = "www.biomart.org"
    path = "/biomart/martservice"
    def __init__(self, name="ensembl", virtualSchema="default", connection=None,
                 database="ensembl_mart_60", default="1",
                 displayName="ENSEMBL GENES 60 (SANGER UK)", host="www.biomart.org",
                 includeDatasets="", martUser="", path="/biomart/martservice",
                 port="80", serverVirtualSchema="default", visible="1", **kwargs):
        
        self.name = name
        self.virtualSchema = virtualSchema
        self.database = database,
        self.default = default
        self.displayName = displayName
        self.host = host
        self.includeDatasets = includeDatasets
        self.martUser = martUser  
        self.path = path
        self.port = port
        self.serverVirtualSchema = serverVirtualSchema
        self.visible = visible
        self.__dict__.update(kwargs.items())
        
        if connection is None:
            connection = BioMartConnection()
            
        if kwargs.get("redirect", None) == "1" and BioMartConnection.FOLLOW_REDIRECTS:
            redirect = BioMartConnection("http://" + self.host + ":" + self.port + self.path, cache=connection.cache)
            try:
                registry = redirect.registry()
                connection = redirect
            except urllib2.HTTPError, ex:
                warnings.warn("'%s' is not responding!, using the default original connection. %s" % (redirect.address, str(ex)))
            
        self.connection = connection
        
#        self.connection = BioMartConnection("http://" + self.host + ":" + self.port + self.path) if connection is None \
#                            or (kwargs.get("redirect", None) == "1" and BioMartConnection.FOLLOW_REDIRECTS) else connection

    @cached    
    def _datasets_index(self):
        keys = ["datasetType", "internalName", "displayName", "visible", "assembly", "_5", "_6", "virtualSchema", "date"]
        
        try:
            datasets = self.connection.datasets(mart=self.name, virtualSchema=self.virtualSchema).read()
        except BioMartError, ex:
            if self.virtualSchema == "default":
                datasets = self.connection.datasets(mart=self.name).read()
            else:
                raise
        datasets = de_tab(datasets)
        return [dict(zip(keys, line)) for line in datasets]
    
    @cached
    def datasets(self):
        """ Return a list of all datasets (BioMartDataset instances) in this database
        """
        return drop_iter(BioMartDataset(mart=self.name, connection=self.connection, **dataset) for dataset in self._datasets_index())
    
    def dataset_attributes(self, dataset, **kwargs):
        """ Return a list of dataset attributes
        """
        dataset = self._find_dataset(dataset)
        return dataset.attributes()
    
    def dataset_filters(self, dataset, **kwargs):
        """ Return a list of dataset filters
        """
        dataset = self._find_dataset(dataset)
        return dataset.filters()
    
    def dataset_query(self, dataset, filters=[], attributes=[]):
        """ Return an dataset query based on dataset, filters and attributes
        """
        dataset = self._find_dataset(dataset)
        return BioMartQuery(self.con, [dataset], filters, attributes).run()
        
    def _find_dataset(self, internalName):
        for dataset in self.datasets():
            if dataset.internalName == internalName:
                return dataset
        raise ValueError("Uknown dataset name '%s'" % internalName)
    
        
    def __getitem__(self, name):
        try:
            return self._find_dataset(name)
        except ValueError:
            raise KeyError(name)
    
    
    def search(self, string, relevance=False):
        strings = string.lower().split()
        results = []
        for dataset, conf in drop_iter((dataset, dataset.configuration()) for dataset in self.datasets()):
            trees = conf.attribute_pages() + conf.attribute_groups() + \
                    conf.attribute_collections() + conf.attributes()
                    
            names = " ".join([getattr(tree, "displayName", "") for tree in trees]).lower()
            count = sum([names.count(s) for s in strings])
            if count:
                results.append((count ,dataset))
        return results
        
class BioMartDataset(object):
    """ An object representing a BioMart dataset (returned by BioMartDatabase)
    """
    
    # these attributes store the result of ".../martservice?type=datasets&mart=..." query  
    FIELDS = ["datasetType", "internalName", "displayName", "visible", "assembly", "_5", "_6", "virtualSchema", "date"]
    
    def __init__(self, mart="ensembl", internalName="hsapiens_gene_ensembl", virtualSchema="default", connection=None, 
                 datasetType="TableSet", displayName="", visible="1", assembly="", date="", **kwargs):
        self.connection = connection if connection is not None else BioMartConnection()
        self.mart = mart
        self.internalName = internalName
        self.virtualSchema = virtualSchema
        self.datasetType = datasetType
        self.displayName = displayName
        self.visible = visible
        self.assembly = assembly
        self.date = date
        self.__dict__.update(kwargs)
        self._attributes = None
        self._filters = None
        
        
    @cached
    def attributes(self):
        """ Return a list of available attributes for this dataset (Attribute instances)
        """
        stream = self.connection.attributes(dataset=self.internalName, virtualSchema=self.virtualSchema)
        response = stream.read()
        lines = response.splitlines()
        return [Attribute(line) for line in lines if line.strip()]
        
    
    @cached
    def filters(self):
        """ Return a list of available filters for this dataset (Filter instances)
        """
        stream = self.connection.filters(dataset=self.internalName, virtualSchema=self.virtualSchema)
        response = stream.read()
        lines = response.splitlines()
        return [Filter(line) for line in lines if line.strip()]
    
    
    @cached
    def configuration(self, parser=None):
        """ Return the configuration tree for this dataset (DatasetConfig instance)
        """
        stream = self.connection.configuration(dataset=self.internalName, virtualSchema=self.virtualSchema)
        response = stream.read()
        doc = parseXML(response, parser)
        config = list(doc.elements("DatasetConfig"))[0]
        return DatasetConfig(BioMartRegistry(self.connection), config.tag, config.attributes, config.children)
        
        
    def get_data(self, attributes=[], filters=[], unique=False):
        """ Constructs and runs a BioMartQuery returning its results 
        """ 
        return BioMartQuery(self.connection, dataset=self, attributes=attributes, filters=filters,
                             uniqueRows=unique, virtualSchema=self.virtualSchema).run()
    
    
    def count(self, filters=[], unique=False):
        """ Constructs and runs a BioMartQuery to count the number of returned lines
        """
        return BioMartQuery(self.connection, dataset=self, filters=filters, uniqueRows=unique,
                            virtualSchema=self.virtualSchema).get_count()
    
                            
    def get_example_table(self, attributes=[], filters=[], unique=False):
        query = BioMartQuery(self.connection, dataset=self, attributes=attributes, filters=filters, uniqueRows=unique,
                            virtualSchema=self.virtualSchema)
        return query.get_example_table()
        

class BioMartQuery(object):
    """ A class for constructing a query to run on a BioMart server
    
    Example::
        >>> BioMartQuery(connection, dataset="hsapiens_gene_ensembl", attributes=["ensembl_transcript_id",
        ...     "chromosome_name"], filters=[("chromosome_name", ["22"])]).get_count()
        1221
        >>> #Equivalent to
        ...
        >>> query = BioMartQuery(connection)
        >>> query.set_dataset("hsapiens_gene_ensembl")
        >>> query.add_filter("chromosome_name", "22")
        >>> query.add_attribute("ensembl_transcript_id")
        >>> query.add_attribute("chromosome_name")
        >>> query.get_count()
        1221
    """
    class XMLQuery(object):
        XML = """<?xml version="1.0" encoding="UTF-8"?> 
<!DOCTYPE Query> 
<Query virtualSchemaName = "%(virtualSchemaName)s" formatter = "%(formatter)s" header = "%(header)s" uniqueRows = "%(uniqueRows)s" count = "%(count)s" datasetConfigVersion = "%(version)s" >  
%(datasets_xml)s 
</Query>"""
        version = "0.6"
        def __init__(self, query):
            self.query = query
            self._query = query._query
            self.__dict__.update(query.__dict__) 
        
        def get_xml(self, count=None, header=False):
            datasets_xml = "\n".join(self.query_dataset(*query) for query in self._query)
            
            links_xml = self.query_links(self._query)
            datasets_xml += "\n" + links_xml
            
            count = self.count if count is None else count 
            args = dict(datasets_xml=datasets_xml,
                        uniqueRows="1" if self.uniqueRows else "0",
                        count="1" if count else "",
                        header= "1" if header else "0",
                        virtualSchemaName=self.virtualSchema,
                        formatter=self.format,
                        version=self.version)
            xml = self.XML % args
            return xml
        
        def query_attributes(self, attributes):
            def name(attr):
                if isinstance(attr, Attribute):
                    return attr.internalName
                elif isinstance(attr, AttributeDescription):
                    return attr.internalName
                else:
                    return str(attr)
                
            xml = "\n".join('\t\t<Attribute name = "%s" />' % name(attr).lower() for attr in attributes)
            return xml
        
        def query_filter(self, filters):
            def name(filter):
                if isinstance(filter, Filter):
                    return filter.internalName
                elif isinstance(filter, FilterDescription):
                    return getattr(filter, "field", getattr(filter, "internalName"))
                else:
                    return str(filter)
                
            def args(args):
                if isinstance(args, list):
                    return 'value="%s"' % ",".join(args)
                elif isinstance(args, basestring):
                    return 'value="%s"' % args
                elif isinstance(args, dict):
                    return " ".join(['%s="%s"' % (key, value) for key, value in args.iteritems()])
                    
#            value = lambda value: str(value) if not isinstance(value, list) else ",".join([str(v) for v in value])
#            xml = "\n".join('\t\t<Filter name = "%s" value="%s" />' % (name(fil), value(val)) for fil, val in filters)
            xml = "\n".join('\t\t<Filter name = "%s" %s />' % (name(fil), args(val)) for fil, val in filters)
            return xml
        
        def query_dataset(self, dataset, attributes, filters):
            name, interface = (dataset.internalName, "default") if \
                    isinstance(dataset, BioMartDataset) else (dataset, "default")
            xml = '\t<Dataset name = "%s" interface = "%s" >\n' % (name, interface)
            xml += self.query_attributes(attributes) + "\n"
            xml += self.query_filter(filters)
            xml += '\n\t</Dataset>'
            return xml
        
        def query_links(self, query):
            def name(dataset):
                if isinstance(dataset, BioMartDataset):
                    return getattr(dataset, "internalName", getattr(dataset, "name", None))
                else:
                    return str(dataset)
                
            if len(query) == 2:
                return '\t<Links source="%s" target="%s"/>' % (name(query[0][0]), name(query[1][0]))
            else:
                return ""
        
    class XMLQueryOld(XMLQuery):
        version = "0.4"
        def query_filter(self, filters):
            def name(filter):
                if isinstance(filter, Filter):
                    return filter.internalName
                elif isinstance(filter, FilterDescription):
                    return getattr(filter, "field", getattr(filter, "internalName"))
                else:
                    return str(filter)
                
            value = lambda value: str(value) if not isinstance(value, list) else ",".join([str(v) for v in value])
            xml = "\n".join('\t\t<ValueFilter name = "%s" value="%s" />' % (name(fil), value(val)) for fil, val in filters)
            return xml
        
    def __init__(self, registry, virtualSchema="default", dataset=None, attributes=[], filters=[], count=False, uniqueRows=False,
                 format="TSV"):
        if isinstance(registry, BioMartConnection):
            self.registry = BioMartRegistry(registry)
            self.virtualSchema = virtualSchema
        elif isinstance(registry, BioMartVirtualSchema):
            self.registry = registry.connection
            self.virtualSchema = registry.name
        else:
            self.registry = registry
            self.virtualSchema = virtualSchema
            
        self._query = []
        if dataset:
            self.set_dataset(dataset)
            for attr in attributes:
                self.add_attribute(attr)
            for filter, value in filters:
                self.add_filter(filter, value)
        self.count = count
        self.uniqueRows = uniqueRows
        self.format = format
        
    def set_dataset(self, dataset):
        self._query.append((dataset, [], []))
    
    def add_filter(self, filter, value):
        self._query[-1][2].append((filter, value))
    
    def add_attribute(self, attribute):
        self._query[-1][1].append(attribute)
    
    def get_count(self):
        count = self.run(count=True)
        if count.strip():
            count = int(count.strip())
        else:
            count = 0
        return count
    
    def run(self, count=None, header=False):
        stream = self.registry.connection.request(query=self.xml_query(count=count, header=header).replace("\n", "").replace("\t", ""))
        stream = stream.read()
        if stream.startswith("Query ERROR:"):
            raise BioMartQueryError(stream)
        return stream
    
    def set_unique(self, unique=False):
        self.uniqueRows = unique
    
    def xml_query(self, count=None, header=False):  
        self.version = version = "0.4"
         
        schema = self.virtualSchema
        
        dataset = self._query[0][0]
        dataset = dataset.internalName if isinstance(dataset, BioMartDataset) else dataset
        
        conf = self.registry.dataset(dataset, virtualSchema=self.virtualSchema).configuration()
        self.version = version = getattr(conf, "softwareVersion", "0.4")
        
        if version > "0.4":
            xml = self.XMLQuery(self).get_xml(count, header)
        else:
            xml = self.XMLQueryOld(self).get_xml(count, header)
        return xml
        
    def get_example_table(self):
        import orange
        data = self.run(count=False, header=True)
        
        if self.format.lower() == "tsv":
            header, data = data.split("\n", 1)
            domain = orange.Domain([orange.StringVariable(name) for name in header.split("\t")], False)
            data = [line.split("\t") for line in data.split("\n") if line.strip()]
            return orange.ExampleTable(domain, data) if data else None
        elif self.format.lower() == "fasta":
            domain = orange.Domain([orange.StringVariable("id"), orange.StringVariable("sequence")], False) #TODO: meaningful id
            examples = []
            from StringIO import StringIO
            from Bio import SeqIO
            for seq in SeqIO.parse(StringIO(data), "fasta"):
                examples.append([seq.id, str(seq.seq)])
            return orange.ExampleTable(domain, examples)
        else:
            raise BioMartError("Unsupported format: %" % self.format)


def get_pointed(self):
    if self.is_pointer():
        pointerDataset = self.pointerDataset
        if hasattr(self, "pointerAttribute"):
            name, getter = self.pointerAttribute, "AttributeDescription"
        elif hasattr(self, "pointerFilter"):
            name, getter = self.pointerFilter, "FilterDescription"
        
        dataset = self.registry.dataset(pointerDataset)
        
        conf = dataset.configuration()
        desc = list(conf.elements(getter, internalName=name))
        if desc:
            return dataset, desc[0]
        else:
            warnings.warn("Could not resolve pointer '%s' in '%s'" % (name, pointerDataset), UserWarning, stacklevel=2)
            return None, None
        
        
def is_pointer(self):
    return hasattr(self, "pointerAttribute") or hasattr(self, "pointerFilter")


class ConfigurationNode(XMLNode):
    def __init__(self, registry, *args, **kwargs):
        XMLNode.__init__(self, *args, **kwargs)
        self.registry = registry
        self.__dict__.update([(self._name_mangle(name), value) \
                               for name, value in self.attributes.iteritems()])
        
        self.children = [self._factory(child) for child in self.children]
        
    def _name_mangle(self, name):
        if name in self.__dict__:
            return name + "_"
        else:
            return name
        
    def _factory(self, node):
        tag = node.tag
        class_ = globals().get(tag, None)
        if not class_:
            raise ValueError("Unknown node '%s;" % tag)
        else:
            return class_(self.registry, tag, node.attributes, node.children)
        
    def elements(self, tag=None, **kwargs):
        if isinstance(tag, type):
            tag = tag.__name__
        return XMLNode.elements(self, tag, **kwargs)
    
    def elements_top(self, tag=None, **kwargs):
        if isinstance(tag, type):
            tag = tag.__name__
        return XMLNode.elements_top(self, tag, **kwargs)
    
    def __repr__(self):
        return '%s("%s", %s)' % (self.__class__.__name__, self.tag, repr(self.attributes))

    
class DatasetConfig(ConfigurationNode):
    pass

class Exportable(ConfigurationNode):
    pass

class Importable(ConfigurationNode):
    pass

class FilterPage(ConfigurationNode):
    pass

class FilterGroup(ConfigurationNode):
    pass

class FilterCollection(ConfigurationNode):
    pass

class FilterDescription(ConfigurationNode):

    is_pointer = is_pointer
    get_pointed = get_pointed
    
class AttributePage(ConfigurationNode):
    pass

class AttributeGroup(ConfigurationNode):
    pass

class AttributeCollection(ConfigurationNode):
    pass

class AttributeDescription(ConfigurationNode):
    
    is_pointer = is_pointer
    get_pointed = get_pointed
    
class Option(ConfigurationNode):
    pass

class PushAction(ConfigurationNode):
    pass

class MainTable(ConfigurationNode):
    pass

class Key(ConfigurationNode):
    pass

    

if __name__ == "__main__":
    con = BioMartConnection("http://www.biomart.org/biomart/martservice")
    registry = BioMartRegistry(con)
    for schema in  registry.virtual_schemas():
        print "Virtual schema '%s'" % schema.name
        for mart in schema.databases():
            print "\tMart: '%s' ('%s'):" % (mart.name, mart.displayName)
            for dataset in mart.datasets():
                print "\t\t Dataset '%s' %s' '%s'" % (dataset.datasetType, dataset.internalName, dataset.displayName)
 
    database = BioMartDatabase(name="dicty", connection=con)
    datasets = database.datasets()
    print datasets
    dataset = datasets[2]
    configuration =  dataset.configuration()
    attr = dataset.attributes()
    print attr
    filters = dataset.filters()
    print filters
    reg = BioMartRegistry(con)
    
    dataset = reg.dataset("scerevisiae_gene_ensembl")
    query = BioMartQuery(con, dataset="scerevisiae_gene_ensembl",
                         attributes=["ensembl_transcript_id", "transcript_exon_intron"],
                         filters = [("chromosome_name", "I"), ("with_wikigene", {"excluded":"1"})],
                         format="FASTA")
    print query.xml_query()
    print query.run()
    
    data = query.get_example_table()
    data.save("seq.tab")
    
    
    import doctest
    doctest.testmod(extraglobs={"connection": con, "registry":registry}, optionflags=doctest.ELLIPSIS)
    



from __future__ import absolute_import

import math
from collections import defaultdict

import scipy.stats

import numpy

import Orange, Orange.utils, statc

from .obiGsea import takeClasses
from .obiAssess import pca, PLSCall, corgs_activity_score
from . import obiExpression, obiGene, obiGeneSets, obiGsea, stats

class GeneSetTrans(object):

    __new__ = Orange.utils._orange__new__(object)

    def _mat_ni(self, data):
        """ With cached gene matchers. """
        if data.domain not in self._cache:
            self._cache[data.domain] = mat_ni(data, self.matcher)
        return self._cache[data.domain]

    def _match_instance(self, instance, geneset, takegenes=None):
        nm, name_ind = self._mat_ni(instance)
        genes = [ nm.umatch(gene) for gene in geneset ]
        if takegenes:
            genes = [ genes[i] for i in takegenes ]
        return nm, name_ind, genes

    def _match_data(self, data, geneset, odic=False):
        nm, name_ind = self._mat_ni(data)
        genes = [ nm.umatch(gene) for gene in geneset ]
        if odic:
            to_geneset = dict(zip(genes, geneset))
        takegenes = [ i for i,a in enumerate(genes) if a != None ]
        genes = [ genes[i] for i in takegenes ]
        if odic:
            return nm, name_ind, genes, takegenes, to_geneset
        else:
            return nm, name_ind, genes, takegenes

    def __init__(self, matcher=None, gene_sets=None, min_size=3, max_size=1000, min_part=0.1, class_values=None):
        self.matcher = matcher
        self.gene_sets = gene_sets
        self.min_size = min_size
        self.max_size = max_size
        self.min_part = min_part
        self.class_values = class_values
        self._cache = {}

    def __call__(self, data, weight_id=None):

        #selection of classes and gene sets
        data = takeClasses(data, classValues=self.class_values)
        nm,_ =  self._mat_ni(data)
        gene_sets = select_genesets(nm, self.gene_sets, self.min_size, self.max_size, self.min_part)

        #build a new domain
        newfeatures = self.build_features(data, gene_sets)
        newdomain = Orange.data.Domain(newfeatures, data.domain.class_var)
        return Orange.data.Table(newdomain, data)

    def build_features(self, data, gene_sets):
        return [ self.build_feature(data, gs) for gs in gene_sets ]


def normcdf(x, mi, st):
    return 0.5*(2. - stats.erfcc((x - mi)/(st*math.sqrt(2))))

class AT_edelmanParametric(object):

    def __init__(self, **kwargs):
        for a,b in kwargs.items():
            setattr(self, a, b)

    def __call__(self, nval):

        if self.mi1 == None or self.mi2 == None or self.st1 == None or self.st2 == None:
            return 0.0 

        val = nval

        try:
            if val >= self.mi1:
                p1 = 1 - normcdf(val, self.mi1, self.st1)
            else:
                p1 = normcdf(val, self.mi1, self.st1)

            if val >= self.mi2:
                p2 = 1 - normcdf(val, self.mi2, self.st2)
            else:
                p2 = normcdf(val, self.mi2, self.st2)

            #print p1, p2
            return math.log(p1/p2)
        except:
            #print p1, p2, "exception"
            return 0

class AT_edelmanParametricLearner(object):
    """
    Returns attribute transfromer for Edelman parametric measure for a
    given attribute in the dataset.  Edelman et al, 06. Modified a bit.
    """

    def __init__(self, a=None, b=None):
        """
        a and b are choosen class values.
        """
        self.a = a
        self.b = b

    def __call__(self, i, data):
        cv = data.domain.classVar
        #print data.domain

        if self.a == None: self.a = cv.values[0]
        if self.b == None: self.b = cv.values[1]

        def avWCVal(value):
            return [ex[i].value for ex in data if ex[-1].value == value and not ex[i].isSpecial() ]

        list1 = avWCVal(self.a)
        list2 = avWCVal(self.b)

        mi1 = mi2 = st1 = st2 = None

        try:
            mi1 = statc.mean(list1)
            st1 = statc.std(list1)
        except:
            pass
    
        try:
            mi2 = statc.mean(list2)
            st2 = statc.std(list2)
        except:
            pass

        return AT_edelmanParametric(mi1=mi1, mi2=mi2, st1=st1, st2=st2)

class AT_loess(object):

    def __init__(self, **kwargs):
        for a,b in kwargs.items():
            setattr(self, a, b)

    def __call__(self, nval):

        val = nval

        def saveplog(a,b):
            try:
                return math.log(a/b)
            except:
                if a < b:
                    return -10
                else:
                    return +10

        try:
            ocene = self.condprob(val)
            if sum(ocene) < 0.01:
                return 0.0
            return saveplog(ocene[0], ocene[1])

        except:
            return 0.0

class AT_loessLearner(object):

    def __call__(self, i, data):
        try:
            ca = Orange.statistics.contingency.VarClass(data.domain.attributes[i], data)
            a =  Orange.statistics.estimate.ConditionalLoess(ca, nPoints=5)
            return AT_loess(condprob=a)
        except:
            return AT_loess(condprob=None)

def nth(l, n):
    return [a[n] for a in l]

class Assess(GeneSetTrans):
    """
    Uses the underlying GSEA code to select genes.
    Takes data and creates attribute transformations.
    """

    def __init__(self, rankingf=None, **kwargs):
        self.rankingf = rankingf
        if self.rankingf == None:
            self.rankingf = AT_edelmanParametricLearner()
        super(Assess, self).__init__(**kwargs)

    def build_features(self, data, gene_sets):

        attributes = []

        #attrans: { i_orig: ranking_function }
        attrans = [ self.rankingf(iat, data) for iat, at in enumerate(data.domain.attributes) ]

        nm_all, _ =  self._mat_ni(data)

        for gs in gene_sets:

            at = Orange.feature.Continuous(name=str(gs))

            geneset = list(gs.genes)
            nm, name_ind, genes, takegenes, to_geneset = self._match_data(data, geneset, odic=True)
            genes = set(genes)
            
            def t(ex, w, geneset=geneset, takegenes=takegenes, nm=nm, attrans=attrans):

                nm2, name_ind2, genes2 = self._match_instance(ex, geneset, takegenes)

                ex_atts = [ at.name for at in ex.domain.attributes ]
                new_atts = [ name_ind[nm.umatch(an)] if nm.umatch(an) != None else None
                    for an in ex_atts ]
                #new_atts: indices of genes in original data for that sample 
                #POSSIBLE REVERSE IMPLEMENTATION (slightly different
                #for data from different chips):
                #save pairs together and sort (or equiv. dictionary transformation)

                indexes = filter(lambda x: x[0] != None, zip(new_atts, range(len(ex_atts))))

                lcor = [ attrans[index_in_data](ex[index_in_ex].value) 
                    for index_in_data, index_in_ex in indexes if
                    ex[index_in_ex].value != '?' ]
                #indexes in original lcor, sorted from higher to lower values
                ordered = obiGsea.orderedPointersCorr(lcor) 
                #subset = list of indices, lcor = correlations, ordered = order
                subset = [ name_ind2[g] for g in genes2 ]
                return obiGsea.enrichmentScoreRanked(subset, lcor, ordered)[0] 

            at.get_value_from = t
            attributes.append(at)

        return attributes
   
def setSig_example_geneset(ex, data):
    """ Gets learning data and example with the same domain, both
    containing only genes from the gene set. """

    distances = [ [], [] ]    

    def pearsonr(v1, v2):
        return numpy.corrcoef([v1, v2])[0,1]

    def pearson(ex1, ex2):
        #leaves undefined elements out

        attrs = range(len(ex1.domain.attributes))
        vals1 = [ ex1[i].value for i in attrs ]
        vals2 = [ ex2[i].value for i in attrs ]

        common = [ True if v1 != "?" and v2 != "?" else False \
            for v1,v2 in zip(vals1,vals2) ]
        vals1 = [ v for v,c in zip(vals1, common) if c ]
        vals2 = [ v for v,c in zip(vals2, common) if c ]

        return numpy.corrcoef([vals1, vals2])[0,1]

    def ttest(ex1, ex2):
        try:
            return stats.lttest_ind(ex1, ex2)[0]
        except:
            return 0.0
    
    #maps class value to its index
    classValueMap = dict( [ (val,i) for i,val in enumerate(data.domain.class_var.values) ])
 
    #create distances to all learning data - save or other class
    for c in data:
        distances[classValueMap[c[-1].value]].append(pearson(c, ex))

    return ttest(distances[0], distances[1])

def mat_ni(data, matcher):
    nm = matcher([at.name for at in data.domain.attributes])
    name_ind = dict((n.name,i) for i,n in enumerate(data.domain.attributes))
    return nm, name_ind

def select_genesets(nm, gene_sets, min_size=3, max_size=1000, min_part=0.1):
    """ Returns a list of gene sets that have sizes in limits """

    def ok_sizes(gs):
        """compares sizes of genesets to limitations"""
        transl = filter(lambda x: x != None, [ nm.umatch(gene) for gene in gs.genes ])
        if len(transl) >= min_size \
            and len(transl) <= max_size \
            and float(len(transl))/len(gs.genes) >= min_part:
            return True
        return False

    return filter(ok_sizes, gene_sets) 

def vou(ex, gn, indices):
    """ returns the value or "?" for the given gene name gn"""
    if gn not in indices:
        return "?"
    else:
        return ex[indices[gn]].value

class SetSig(GeneSetTrans):

    def build_feature(self, data, gs):

        at = Orange.feature.Continuous(name=str(gs))

        def t(ex, w, gs=gs, data=data): #copy od the data
            geneset = list(gs.genes)

            nm, name_ind, genes, takegenes = self._match_data(data, geneset)
            nm2, name_ind2, genes2 = self._match_instance(ex, geneset, takegenes)

            domain = Orange.data.Domain([data.domain.attributes[name_ind[gene]] for gene in genes], data.domain.class_var)
            datao = Orange.data.Table(domain, data)
           
            #convert the example to the same domain
            exvalues = [ vou(ex, gn, name_ind2) for gn in genes2 ] + [ "?" ]
            example = Orange.data.Instance(domain, exvalues)

            return setSig_example_geneset(example, datao) #only this one is setsig specific
     
        at.get_value_from = t
        return at

class ParametrizedTransformation(GeneSetTrans):

    def _get_par(self, datao):
        pass
        
    def _use_par(self, ex, constructt):
        pass
    
    def build_feature(self, data, gs):

        at = Orange.feature.Continuous(name=str(gs))

        geneset = list(gs.genes)
        nm, name_ind, genes, takegenes = self._match_data(data, geneset)
        domain = Orange.data.Domain([data.domain.attributes[name_ind[gene]] for gene in genes], data.domain.class_var)
        datao = Orange.data.Table(domain, data)

        constructt = self._get_par(datao)

        def t(ex, w, geneset=geneset, constructt=constructt, takegenes=takegenes, domain=domain):
            nm2, name_ind2, genes2 = self._match_instance(ex, geneset, takegenes)
          
            #convert the example to the same domain
            exvalues = [ vou(ex, gn, name_ind2) for gn in genes2 ] + [ "?" ]
            ex = numpy.array(exvalues[:-1])

            return self._use_par(ex, constructt)
            
        at.get_value_from = t
        return at

class PLS(ParametrizedTransformation):

    def _get_par(self, datao):
        return PLSCall(datao, nc=1, y=[datao.domain.class_var])
        
    def _use_par(self, ex, constructt):
        xmean, W, P, _ = constructt
        ex = ex - xmean # same input transformation

        nc = W.shape[1]

        TR = numpy.empty((1, nc))
        XR = ex

        dot = numpy.dot

        for i in range(nc):
           t = dot(XR, W[:,i].T)
           XR = XR - t*numpy.array([P[:,i]])
           TR[:,i] = t

        return TR[0][0]
        
class PCA(ParametrizedTransformation):

    def _get_par(self, datao):
        return pca(datao)

    def _use_par(self, arr, constructt):
        evals, evect, xmean = constructt

        arr = arr - xmean # same input transformation - a row in a matrix
        ev0 = evect[0] #this is a row in a matrix - do a dot product
        a = numpy.dot(arr, ev0)

        return a

class SimpleFun(GeneSetTrans):

    def build_feature(self, data, gs):

        at = Orange.feature.Continuous(name=str(gs))

        def t(ex, w, gs=gs):
            geneset = list(gs.genes)
            nm2, name_ind2, genes2 = self._match_instance(ex, geneset)
           
            exvalues = [ vou(ex, gn, name_ind2) for gn in genes2 ] + [ "?" ]
            exvalues = filter(lambda x: x != "?", exvalues)

            return self.fn(exvalues)
     
        at.get_value_from = t
        return at

class Mean(SimpleFun):

    def __init__(self, **kwargs):
       self.fn = numpy.mean
       super(Mean, self).__init__(**kwargs)

class Median(SimpleFun):

    def __init__(self, **kwargs):
       self.fn = numpy.median
       super(Median, self).__init__(**kwargs)

class GSA(GeneSetTrans):

    def build_features(self, data, gene_sets):

        attributes = []

        def tscorec(data, at, cache=None):
            ma = obiExpression.MA_t_test()(at,data)
            return ma

        tscores = [ tscorec(data, at) for at in data.domain.attributes ]

        def to_z_score(t):
            return float(scipy.stats.norm.ppf(scipy.stats.t.cdf(t, len(data)-2)))

        zscores = map(to_z_score, tscores)

        for gs in gene_sets:

            at = Orange.feature.Continuous(name=str(gs))

            geneset = list(gs.genes)
            nm, name_ind, genes, takegenes, to_geneset = self._match_data(data, geneset, odic=True)
            #take each gene only once
            genes = set(genes)

            D = numpy.mean([max(zscores[name_ind[g]],0) for g in genes]) \
                + numpy.mean([min(zscores[name_ind[g]],0) for g in genes])

            if D >= 0:
                consider_genes = [ to_geneset[g] for g in genes if zscores[name_ind[g]] > 0.0 ]
            else:
                consider_genes = [ to_geneset[g] for g in genes if zscores[name_ind[g]] < 0.0 ]

            def t(ex, w, consider_genes=consider_genes):
                nm2, name_ind2, genes2 = self._match_instance(ex, consider_genes)
              
                #convert the example to the same domain
                exvalues = [ vou(ex, gn, name_ind2) for gn in genes2 ] + [ "?" ]
                exvalues = filter(lambda x: x != "?", exvalues)
              
                return numpy.mean(exvalues)

            at.get_value_from = t
            attributes.append(at)

        return attributes

def tscorec(data, at, cache=None):
    """ Cached attribute  tscore calculation """
    if cache != None and at in cache: return cache[at]
    ma = obiExpression.MA_t_test()(at,data)
    if cache != None: cache[at] = ma
    return ma

def nth(l, n):
    return [a[n] for a in l]

def compute_corg(data, inds, tscorecache):
    """
    Compute CORG for this geneset specified with gene inds
    in the example table. Output is the list of gene inds
    in CORG.

    """
    #order member genes by their t-scores: decreasing, if av(t-score) >= 0,
    #else increasing
    tscores = [ tscorec(data, at, tscorecache) for at in inds ]
    sortedinds = nth(sorted(zip(inds,tscores), key=lambda x: x[1], \
        reverse=numpy.mean(tscores) >= 0), 0)

    def S(corg):
        """ Activity score separation - S(G) in 
        the article """
        asv = Orange.feature.Continuous(name='AS')
        asv.getValueFrom = lambda ex,rw: Orange.data.Value(asv, corgs_activity_score(ex, corg))
        data2 = Orange.data.Table(Orange.data.Domain([asv], data.domain.classVar), data)
        return abs(tscorec(data2, 0)) #FIXME absolute - nothing in the article abs()
            
    #greedily find CORGS procing the best separation
    g = S(sortedinds[:1])
    bg = 1
    for a in range(2, len(sortedinds)+1):
        tg = S(sortedinds[:a])
        if tg > g:
            g = tg
            bg = a
        else:
            break
        
    return sortedinds[:a]

class CORGs(ParametrizedTransformation):
    """
    WARNING: input has to be z_ij table! each gene needs to be normalized
    (mean=0, stdev=1) for all samples.
    """

    def __call__(self, *args, **kwargs):
        self.tscorecache = {} #reset a cache
        return super(CORGs, self).__call__(*args, **kwargs)

    def build_feature(self, data, gs):

        at = Orange.feature.Continuous(name=str(gs))
        geneset = list(gs.genes)

        nm, name_ind, genes, takegenes, to_geneset = self._match_data(data, geneset, odic=True)
        indices = compute_corg(data, [ name_ind[g] for g in genes ], self.tscorecache)

        ind_names = dict( (a,b) for b,a in name_ind.items() )
        selected_genes = sorted(set([to_geneset[ind_names[i]] for i in indices]))
            
        def t(ex, w, corg=selected_genes): #copy od the data
            nm2, name_ind2, genes2 = self._match_instance(ex, corg, None)
            exvalues = [ vou(ex, gn, name_ind2) for gn in genes2 ]
            return sum(v if v != '?' else 0.0 for v in exvalues)/len(corg)**0.5
     
        at.get_value_from = t
        return at

if __name__ == "__main__":

    data = Orange.data.Table("iris")
    gsets = obiGeneSets.collections({
        #"ALL": ['sepal length', 'sepal width', 'petal length', 'petal width'],
        "f3": ['sepal length', 'sepal width', 'petal length'],
        "l3": ['sepal width', 'petal length', 'petal width'],
        })
    matcher = obiGene.matcher([])
    choosen_cv = ["Iris-setosa", "Iris-versicolor"]

    """
    data = Orange.data.Table("DLBCL_200a")
    gsets = obiGeneSets.collections((("KEGG",),"9606"))
    matcher = obiGene.matcher([obiGene.GMKEGG("hsa")])
    choosen_cv = None
    """

    def to_old_dic(d, data):
        ar = defaultdict(list)
        for ex1 in data:
            ex = d(ex1)
            for a,x in zip(d.attributes, ex):
                ar[a.name].append(x.value)
        return ar

    def pp2(ar):
        ol =  sorted(ar.items())
        print '\n'.join([ a + ": " +str(b) for a,b in ol])

    ass = Assess(data, matcher=matcher, gene_sets=gsets, class_values=choosen_cv, min_part=0.0)
    ar = to_old_dic(ass.domain, data[:5])
    pp2(ar)



"""
<name>dictyExpress</name>
<description>Access to data in dictyExpress database.</description>
<icon>icons/dictyExpress.png</icon>
<priority>40</priority>
"""

from __future__ import absolute_import

from collections import defaultdict
import sys

from Orange.orng import orngEnviron
from Orange.OrangeWidgets import OWGUI
from Orange.OrangeWidgets.OWWidget import *

from .. import obiDicty

class MyTreeWidgetItem(QTreeWidgetItem):
    def __contains__(self, text):
        return any(text.upper() in str(self.text(i)).upper() for i in range(self.columnCount()))    

#set buffer file
bufferpath = os.path.join(orngEnviron.directoryNames["bufferDir"], "dicty")
try:
    os.makedirs(bufferpath)
except:
    pass
bufferfile = os.path.join(bufferpath, "database.sq3")

class OWDicty(OWWidget):
    settingsList = ["serverToken", "platform", "selectedExperiments", "server", "buffertime", "excludeconstant" ]
    def __init__(self, parent=None, signalManager=None, name="dictyExpress"):
        OWWidget.__init__(self, parent, signalManager, name)
        self.outputs = [("Example table", ExampleTable)]
        self.serverToken = ""
        self.server = obiDicty.defaddress

        self.platform = None

        self.selectedExperiments = []
        self.buffer = obiDicty.BufferSQLite(bufferfile)

        self.searchString = ""
        self.excludeconstant = False
        
        box = OWGUI.widgetBox(self.controlArea, "Cache")
        OWGUI.button(box, self, "Clear cache", callback=self.clear_buffer)

        OWGUI.checkBox(self.controlArea, self, "excludeconstant", "Exclude labels with constant values" )

        OWGUI.button(self.controlArea, self, "&Commit", callback=self.Commit)
        box  = OWGUI.widgetBox(self.controlArea, "Server")
        OWGUI.lineEdit(box, self, "serverToken","Token", callback=self.Connect)
        OWGUI.rubber(self.controlArea)

        OWGUI.lineEdit(self.mainArea, self, "searchString", "Search", callbackOnType=True, callback=self.SearchUpdate)
        self.experimentsWidget = QTreeWidget()
        self.experimentsWidget.setHeaderLabels(["Strain", "Treatment", "Growth condition", "Platform", "N", "Chips"]) 
        self.experimentsWidget.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.experimentsWidget.setRootIsDecorated(False)
        self.experimentsWidget.setSortingEnabled(True)
##        self.experimentsWidget.setAlternatingRowColors(True)

        self.mainArea.layout().addWidget(self.experimentsWidget)

        self.loadSettings()
        self.dbc = None        

        QTimer.singleShot(0, self.UpdateExperiments)        

        self.resize(800, 600)

    def __updateSelectionList(self, oldList, oldSelection, newList):
        oldList = [oldList[i] for i in oldSelection]
        return [ i for i, new in enumerate(newList) if new in oldList]
    
    def Connect(self):

        address = self.server + "?"
        if self.serverToken:
            address += "token="+self.serverToken+"&"
        try:
            #obiDicty.verbose = 1
            self.dbc = obiDicty.DatabaseConnection(address, buffer=self.buffer)
        except Exception, ex:
            from traceback import print_exception
            print_exception(*sys.exc_info())
            self.error(0, "Error connecting to server" + str(ex))
            return
        self.error(0)

    def clear_buffer(self):
        self.buffer.clear()
        self.UpdateExperiments()

    def UpdateExperiments(self):
        self.chipsl = []
        self.experimentsWidget.clear()
        self.items = []

        if not self.dbc:
            self.Connect()

        """
        self.progressBarInit()

        strains = self.dbc.annotationOptions("sample")["sample"]

        for i, strain in enumerate(strains):
            chips = self.dbc.search("norms", sample=strain)
            annotations = self.dbc.annotations("norms", chips)

            elements = []

            for chip,annot in zip(chips,annotations):
                d = dict(annot)
                elements.append(((d.get("treatment", ""), d.get("growthCond", ""), d.get("platform", "")),chip))

            def different_chips(li):
                # Returns a map, where keys are different elements in li and values their chip ids
                dc = defaultdict(list)
                for a,chip in li:
                    dc[a].append(chip)
                return dc

            typeswchips = different_chips(elements) #types with counts

            for (treatment, cond, platforma),cchips in typeswchips.items():
                self.chipsl.append(cchips)
                num = len(cchips)
                experiment = [strain, treatment, cond, platform, str(num), ','.join(cchips)] 
                self.items.append(MyTreeWidgetItem(self.experimentsWidget, experiment))

            self.progressBarSet((100.0 * i) / len(strains))
        """

        annotations = self.dbc.annotations("norms", None)

        elements = []

        for chip,annot in annotations:
            d = dict(annot)
            elements.append(((d.get("treatment", ""), d.get("growthCond", ""), d.get("platform", ""), d.get("sample", "")),chip))

        def different_chips(li):
            # Returns a map, where keys are different elements in li and values their chip ids
            dc = defaultdict(list)
            for a,chip in li:
                dc[a].append(chip)
            return dc

        typeswchips = different_chips(elements) #types with counts

        for (treatment, cond, platform, strain),cchips in typeswchips.items():
            self.chipsl.append(cchips)
            num = len(cchips)
            experiment = [strain, treatment, cond, platform, str(num), ','.join(cchips)] 
            self.items.append(MyTreeWidgetItem(self.experimentsWidget, experiment))

        for i in range(5):
            self.experimentsWidget.resizeColumnToContents(i)

        self.progressBarFinished()

    def SearchUpdate(self, string=""):
        for item in self.items:
            item.setHidden(not all(s in item for s in self.searchString.split()))

    def Commit(self):
        if not self.dbc:
            self.Connect()
        allTables = []

        import time
        start = time.time()

        pb = OWGUI.ProgressBar(self, iterations=1000)

        table = None

        ids = []
        for item in self.experimentsWidget.selectedItems():
            ids += str(item.text(5)).split(",")

        table = self.dbc.get_single_data(ids=ids, callback=pb.advance, exclude_constant_labels=self.excludeconstant)

        end = int(time.time()-start)
        
        pb.finish()

        #self.send("Example table", None)
#        table.taxid = "352472"
#        table.genesinrows = False
        
        from Orange.orng.orngDataCaching import data_hints
        data_hints.set_hint(table, "taxid", "352472", 10.0)
        data_hints.set_hint(table, "genesinrows", False, 10.0)
        
        self.send("Example table", table)

if __name__ == "__main__":
    app  = QApplication(sys.argv)
##    from pywin.debugger import set_trace
##    set_trace()
    w = OWDicty()
    w.show()
    app.exec_()
    w.saveSettings()
            
        



from __future__ import absolute_import

import cPickle, gzip, os.path, re

import orange
from Orange.orng import orngMisc, orngServerFiles

from . import obiData, obiTaxonomy

def spots_mean(x):
    vs = [v for v in x if v and v<>"?"]
    if len(vs) == 0: return "?"
    return sum(vs)/len(vs)
def spots_median(x):
    vs = [v for v in x if v and v<>"?"]
    if len(vs) == 0: return "?"
    if len(vs) % 2:
        return sorted(vs)/(len(vs)/2)
    else:
        z = sorted(x)
        return (z[len(vs)/2-1] + z[len(vs)/2]) / 2. 
    return sum(vs)/len(vs)
def spots_min(x):
    vs = [v for v in x if v and v<>"?"]
    if len(vs) == 0: return "?"
    return min(vs)/len(vs)
def spots_max(x):
    vs = [v for v in x if v and v<>"?"]
    if len(vs) == 0: return "?"
    return max(vs)/len(vs)

p_assign = re.compile(" = (.*$)")
p_tagvalue = re.compile("![a-z]*_([a-z_]*) = (.*)$")    
tagvalue = lambda x: p_tagvalue.search(x).groups()

DOMAIN = "GEO"
GDS_INFO_FILENAME = "gds_info.pickled"
FTP_NCBI = "ftp.ncbi.nih.gov"
FTP_DIR = "pub/geo/DATA/SOFT/GDS/"

class GDSInfo:
    def __init__(self, force_update=False):
        path = orngServerFiles.localpath(DOMAIN, GDS_INFO_FILENAME)
        if not os.path.exists(path) or force_update:
            orngServerFiles.download(DOMAIN, GDS_INFO_FILENAME)
        f = file(path, "rb")
        self.info, self.excluded = cPickle.load(f)
    def keys(self): return self.info.keys()
    def items(self): return self.info.items()
    def values(self): return self.info.values()
    def clear(self): return self.info.clear()
    def __getitem__(self, key): return self.info[key]
    def __setitem__(self, key, item): self.info[key] = item
    def __len__(self): return len(self.info)
    def __iter__(self): return iter(self.info)
    def __contains__(self, key): return key in self.info
    

class GeneData:
    """Store mapping between spot id and gene."""
    def __init__(self, spot_id, gene_name, d):
        self.spot_id = spot_id
        self.gene_name = gene_name
        self.data = d

class GDS():
    """GEO DataSet class: read GEO datasets and convert them to ExampleTable."""
    def __init__(self, gdsname, verbose=False, force_download=False):
        self.gdsname = gdsname
        self.verbose = verbose
        self.force_download = force_download
        self.filename = orngServerFiles.localpath(DOMAIN, self.gdsname + ".soft.gz")
        self._getinfo() # to get the info
        taxid = obiTaxonomy.search(self.info["sample_organism"], exact=True)
        self.info["taxid"] = taxid[0] if len(taxid)==1 else None
        self._getspotmap() # to get gene->spot and spot->gene mapping
        self.genes = self.gene2spots.keys()
        self.info["gene_count"] = len(self.genes)
        self.gdsdata = None
        self.data = None
        
    def _download(self):
        """Download GDS data file if not in local cache or forced download requested."""
        localpath = orngServerFiles.localpath(DOMAIN)
        if self.force_download or not os.path.exists(self.filename):
            ftp = obiData.FtpDownloader(FTP_NCBI, localpath, FTP_DIR)
            ftp.retrieve(self.gdsname + ".soft.gz", progressCallback=orngMisc.ConsoleProgressBar()
                         if self.verbose else None)
            if self.verbose:
                # print "Downloading %s" % self.gdsname
                print

    def _getinfo(self):
        """Parse GDS data file and return a dictionary with info."""
        getstate = lambda x: x.split(" ")[0][1:] 
        getid = lambda x: x.rstrip().split(" ")[2]
        self._download()
        f = gzip.open(self.filename)
        state = None; previous_state = None
    
        info = {"subsets" : []}
        subset = None
    
        # GDS information part
        for line in f:
            if line[0] == "^":
                previous_state = state; state = getstate(line)
                if state == "SUBSET":
                    if subset:
                        info["subsets"] += [subset]
                    subset = {"id" : getid(line)}
                if state == "DATASET":
                    info["dataset_id"] = getid(line)
                continue
            if state == "DATASET":
                if previous_state == "DATABASE":
                    tag, value = tagvalue(line)
                    info[tag] = value
                else:
                    if subset:
                        info["subsets"] += [subset]
                    break
            if state == "SUBSET":
                tag, value = tagvalue(line)
                if tag == "description" or tag == "type":
                    subset[tag] = value
                if tag == "sample_id":
                    subset[tag] = value.split(",")
        for t,v in info.items():
            if "count" in t:
                info[t] = int(v)
                
        # sample information
        state = None
        for line in f:
            if state == "header":
                info["samples"] = line.rstrip().split("\t")[2:]
                break
            if line.startswith("!dataset_table_begin"):
                state = "header"

        self.info = info

    def _getspotmap(self, include_spots=None):
        """Return gene to spot and spot to genes mapings."""
        f = gzip.open(self.filename)
        for line in f:
            if line.startswith("!dataset_table_begin"):
                break
        f.readline() # skip header
        spot2gene = {}
        gene2spots = {}
        for line in f:
            if line.startswith("!dataset_table_end"):
                break
            spot, gene = line.rstrip().split("\t")[0:2]
            if include_spots and (spot not in include_spots):
                continue 
            spot2gene[spot] = gene
            gene2spots[gene] = gene2spots.get(gene, []) + [spot]
    
        self.spot2gene = spot2gene
        self.gene2spots = gene2spots
        
    def sample_annotations(self, sample_type=None):
        """Return a dictionary with sample annotation."""
        annotation = {}
        for info in self.info["subsets"]:
            if sample_type and info["type"]<>sample_type:
                continue
            for id in info["sample_id"]:
                annotation.setdefault(id, {})[info["type"]]=info["description"]
        return annotation

    def sample_to_class(self, sample_type=None, missing_class_value=None):
        """Return class values for GDS samples."""
        annotations = self.sample_annotations(sample_type)
        return dict([(sample, "|".join([a for t,a in ann.items()])) for sample, ann in annotations.items()])
        
    def sample_types(self):
        """Return a set of sample types."""
        return set([info["type"] for info in self.info["subsets"]])
    
    def _parse_soft(self, remove_unknown=None):
        """Parse GDS data, returns data dictionary."""
        f = gzip.open(self.filename)
        mfloat = lambda x: float(x) if x<>'null' else '?'
    
        data = {}
        # find the start of the data part
        for line in f:
            if line.startswith("!dataset_table_begin"):
                break
        f.readline()
        
        # load data
        for line in f:
            if line.startswith("!dataset_table_end"):
                break
            d = line.rstrip().split("\t")
            if remove_unknown and (float(d[2:].count('null')) / len(d[2:]) > remove_unknown):
#                print "ZZZ", float(d[2:].count('null')) / len(d[2:]), d[2:]
                continue 
            data[d[0]] = GeneData(d[0], d[1], [mfloat(v) for v in d[2:]])
        self.gdsdata = data
    
    def _to_ExampleTable(self, report_genes=True, merge_function=spots_mean,
                                sample_type=None, missing_class_value=None, transpose=False):
        """Convert parsed GEO format to orange, save by genes or by spots."""
        orng_data = []
        if transpose: # samples in rows
            sample2class = self.sample_to_class(sample_type, missing_class_value)
            cvalues = list(set(sample2class.values()))
            if None in cvalues:
                cvalues.remove(None)
            classvar = orange.EnumVariable(name=sample_type or "class", values=cvalues)
            if report_genes: # save by genes
                atts = [orange.FloatVariable(name=gene) for gene in self.gene2spots.keys()]
                domain = orange.Domain(atts, classvar)
                for (i, sampleid) in enumerate(self.info["samples"]):
                    vals = [merge_function([self.gdsdata[spot].data[i] \
                            for spot in self.gene2spots[gene]]) for gene in self.gene2spots.keys()]
                    orng_data.append(vals + [sample2class[sampleid]])
                
            else: # save by spots
                spots = self.spot2gene.keys()
                atts = [orange.FloatVariable(name=id) for id in spots]
                domain = orange.Domain(atts, classvar)
                for (i, sampleid) in enumerate(self.info["samples"]):
                    orng_data.append([self.gdsdata[spot].data[i] for spot in spots] + [sample2class[sampleid]])
            if missing_class_value == None:
                orng_data = [example for example in orng_data if example[-1] != None]
    
            return orange.ExampleTable(domain, orng_data)
    
        else: # genes in rows
            annotations = self.sample_annotations(sample_type)
            atts = [orange.FloatVariable(name=ss) for ss in self.info["samples"]]
            for i, a in enumerate(atts):
                a.setattr("attributes", annotations[self.info["samples"][i]])
            domain  = orange.Domain(atts, False)
    
            if report_genes: # save by genes
                domain.addmeta(orange.newmetaid(), orange.StringVariable("gene"))
                for g in self.gene2spots.keys():
                    orng_data.append(map(lambda *x: merge_function(x),
                                         *[self.gdsdata[spot].data for spot in self.gene2spots[g]]))
            else: # save by spots
                domain.addmeta(orange.newmetaid(), orange.StringVariable("spot"))
                spots = self.spot2gene.keys()
                orng_data = [self.gdsdata[spot].data for spot in spots]
    
            data = orange.ExampleTable(domain, orng_data)

            if report_genes:
                for i, g in enumerate(self.gene2spots.keys()):
                    data[i]["gene"] = g
            else:
                for i, s in enumerate(spots):
                    data[i]["spot"] = s
            return data
        
    def getdata(self, report_genes=True, merge_function=spots_mean,
                 sample_type=None, missing_class_value=None,
                 transpose=False, remove_unknown=None):
        """Load GDS data and returns a corresponding orange data set,
        spot<->gene mappings and subset info."""
        if self.verbose: print "Reading data ..."
#        if not self.gdsdata:
        self._parse_soft(remove_unknown = remove_unknown)
#        if remove_unknown:
            # some spots were filtered out, need to revise spot<>gene mappings
        self._getspotmap(include_spots=set(self.gdsdata.keys()))
        if self.verbose: print "Converting to example table ..."
        self.data = self._to_ExampleTable(merge_function=merge_function,
                                          sample_type=sample_type, transpose=transpose,
                                          report_genes=report_genes)
        return self.data

    def __str__(self):
        return "%s (%s), samples=%s, features=%s, genes=%s, subsets=%d" % \
              (self.info["dataset_id"],
               self.info["sample_organism"],
               self.info['sample_count'],
               self.info['feature_count'],
               self.info['gene_count'],
               len(self.info['subsets']),
               )


def _float_or_na(x):
    if x.isSpecial():
        return "?"
    return float(x)

def transpose_class_to_labels(data, attcol="sample"):
    """Converts data with genes as attributes to data with genes in rows."""
    if attcol in [v.name for v in data.domain.getmetas().values()]:
        atts = [orange.FloatVariable(str(d[attcol])) for d in data]
    else:
        atts = [orange.FloatVariable("S%d" % i) for i in range(len(data))]
    for i, d in enumerate(data):
        atts[i].setattr("class", str(d.getclass()))
    domain = orange.Domain(atts, False)
    
    newdata = []
    for a in data.domain.attributes:
        newdata.append([_float_or_na(d[a]) for d in data])

    gene = orange.StringVariable("gene")
    id = orange.newmetaid()
    new = orange.ExampleTable(domain, newdata)
    new.domain.addmeta(id, gene)
    for i, d in enumerate(new):
        d[gene] = data.domain.attributes[i].name

    return new

def transpose_labels_to_class(data, class_label=None, gene_label="gene"):
    """Converts data with genes in rows to data with genes as attributes."""
    # if no class_label (attribute type) given, guess it from the data
    if not class_label:
        l = []
        for a in data.domain.attributes:
            l.extend(a.attributes.keys())
        l = list(set(l))
        class_label = l[0]
        if len(set(l)) > 1:
            import warnings
            warnings.warn("More than single attribute label types (%s), took %s"
                          % (", ".join(l), class_label))

    if gene_label in [v.name for v in data.domain.getmetas().values()]:
        atts = [orange.FloatVariable(str(d[gene_label])) for d in data]
    else:
        atts = [orange.FloatVariable("A%d" % i) for i in range(len(data))]
        
    classvalues = list(set([a.attributes[class_label] for a in data.domain.attributes]))
    
    if all(map(lambda x: isinstance(x, (int, long, float, complex)), classvalues)):
        classvar = orange.FloatVariable(class_label)
    else:
        classvar = orange.EnumVariable(class_label, values=classvalues)
        
    domain = orange.Domain(atts + [classvar])
    
    newdata = []
    for a in data.domain.attributes:
        newdata.append([_float_or_na(d[a]) for d in data] + [a.attributes[class_label]])

    sample = orange.StringVariable("sample")
    id = orange.newmetaid()
    new = orange.ExampleTable(domain, newdata)
    new.domain.addmeta(id, sample)
    for i, d in enumerate(new):
        d[sample] = data.domain.attributes[i].name

    return new

def transpose(data):
    """Transposes data matrix, converts class information to attribute label and back"""
    if data.domain.classVar:
        return transpose_class_to_labels(data)
    else:
        return transpose_labels_to_class(data)



"""
KEGG Pathway (from kgml file)

"""
from __future__ import absolute_import

import os
import urllib2

import xml.dom
import xml.parsers
from xml.dom import minidom

from contextlib import closing

from . import conf
from . import caching

from Orange.utils import deprecated_attribute
    
def cached_method(func, cache_name="_cached_method_cache", store=None):
    def wrapper(self, *args, **kwargs):
        sig = (func.__name__,) + args + tuple(sorted(kwargs.items()))
        if not hasattr(self, cache_name):
            setattr(self, cache_name, store() if store is not None else {})
        if sig not in getattr(self, cache_name):
            getattr(self, cache_name)[sig] = func(self, *args, **kwargs)
        return getattr(self, cache_name)[sig]
    return wrapper

class Pathway(object):
    KGML_URL_FORMAT = "http://www.genome.jp/kegg-bin/download?entry={pathway_id}&format=kgml"
    
    def __init__(self, pathway_id, local_cache=None, connection=None):
        if pathway_id.startswith("path:"):
            _, pathway_id = pathway_id.split(":", 1)
            
        self.pathway_id = pathway_id
        if local_cache is None:
            local_cache = conf.params["cache.path"]
        self.local_cache = local_cache
        self.connection = connection
        
    def cache_store(self):
        caching.touch_path(self.local_cache)
        return caching.Sqlite3Store(os.path.join(self.local_cache,
                                                 "pathway_store.sqlite3"))
        
    def _open_last_modified_store(self):
        import shelve
        caching.touch_dir(self.local_cache)
        return caching.Sqlite3Store(os.path.join(self.local_cache,
                                                 "last_modified.sqlite3"))
        
    def _get_kgml(self):
        """ Return an open kgml file for the pathway.
        """
        from datetime import datetime, timedelta
        valid = False
        local_filename = os.path.join(self.local_cache, self.pathway_id + ".xml")
        if os.path.exists(local_filename):
            mtime = os.stat(local_filename).st_mtime
            mtime = datetime.fromtimestamp(mtime)
            now = datetime.now()
            if conf.params["cache.invalidate"] == "always":
                valid = False
            elif conf.params["cache.invalidate"] == "session":
                valid = (now - mtime) < (now - caching._SESSION_START)
            elif conf.params["cache.invalidate"] == "daily":
                valid = (now - mtime) < timedelta(1)
            elif conf.params["cache.invalidate"] == "weekly":
                valid = (now - mtime) < timedelta(7)
            else:
                valid = False
        
        if not valid:
            url = self.KGML_URL_FORMAT.format(pathway_id=self.pathway_id)
            s = urllib2.urlopen(url)
            contents = s.read()
            
            with open(local_filename, "wb") as f:
                f.write(contents)
                
        return open(local_filename, "rb")
    
    def _get_image_filename(self):
        """ Return a filename of a local copy of the pathway image
        """
        # TODO: keep-alive (using httplib if it supports it)
        # better to move all code to use requests package
        
        url = str(self.image)
        
        local_filename = os.path.join(self.local_cache, self.pathway_id + ".png")
        
        if not os.path.exists(local_filename):
            response = urllib2.urlopen(url)
            modified_since = response.headers.get("last-modified")
            image = response.read()
        else:
            request = urllib2.Request(url)
            modified_cache = os.path.join(self.local_cache, "http-last-modified-cache.store")
            with closing(self._open_last_modified_store()) as store:
                modified_since = store.get(url, None)
                
            request.add_header("If-Modified-Since", modified_since)
            try:
                response = urllib2.urlopen(request)
            except urllib2.HTTPError, ex:
                if ex.code == 304:
                    return local_filename
                else:
                    raise
            modified_since = response.headers.get("last-modified")
            image = response.read()
                
        with open(local_filename, "wb") as f:
            f.write(image)
            
        with closing(self._open_last_modified_store()) as store:
            store[url] = modified_since
            
        return local_filename 
        
    def _local_kgml_filename(self):
        """ Return the local kgml xml filename for the pathway.
        """
        local_filename = os.path.join(self.local_cache, self.pathway_id + ".xml")
        return local_filename
        
    class entry(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.graphics = ()
            self.components = []
            self.graphics = dict(dom_element.getElementsByTagName("graphics")[0].attributes.items())
            self.components = [node.getAttribute("id") for node in dom_element.getElementsByTagName("component")]
            
    class reaction(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.substrates = [node.getAttribute("name") for node in dom_element.getElementsByTagName("substrate")]
            self.products = [node.getAttribute("name") for node in dom_element.getElementsByTagName("product")]
            
    class relation(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.subtypes = [node.attributes.items() for node in dom_element.getElementsByTagName("subtype")]
        
    @cached_method
    def pathway_attributes(self):
        return dict(self.pathway_dom().attributes.items())
    
    @property
    def name(self):
        return self.pathway_attributes().get("name")
    
    @property
    def org(self):
        return self.pathway_attributes().get("org")
    
    @property
    def number(self):
        return self.pathway_attributes().get("number")
    
    @property    
    def title(self):
        return self.pathway_attributes().get("title")
    
    @property
    def image(self):
        return self.pathway_attributes().get("image")
    
    @property
    def link(self):
        return self.pathway_attributes().get("link")
    
    @cached_method
    def pathway_dom(self):
        try:
            return minidom.parse(self._get_kgml()).getElementsByTagName("pathway")[0]
        except xml.parsers.expat.ExpatError:
            # TODO: Should delete the cached xml file.
            return None
    
    @cached_method
    def entries(self):
        dom = self.pathway_dom()
        if dom:
            return [self.entry(e) for e in dom.getElementsByTagName("entry")]
        else:
            return []
    
    entrys = deprecated_attribute("entrys", "entries")
    
    @cached_method
    def reactions(self):
        dom = self.pathway_dom()
        if dom:
            return [self.reaction(e) for e in dom.getElementsByTagName("reaction")]
        else:
            return []
    
    @cached_method
    def relations(self):
        dom = self.pathway_dom()
        if dom:
            return [self.relation(e) for e in dom.getElementsByTagName("relation")]
        else:
            return []
    
    def __iter__(self):
        """ Iterate over all elements in the pathway
        """
        return iter(self.all_elements())
    
    def __contains__(self, element):
        """ Retrurn true if element in the pathway
        """
        return element in self.all_elements()
    
    @classmethod
    def split_pathway_id(cls, id):
        path, id = id.split(":") if ":" in id else ("path", id)
        org, id = id[:-5], id[-5:]
        return path, org, id  
    
    @cached_method
    def all_elements(self):
        """ Return all elements
        """
        return reduce(list.__add__, [self.genes(), self.compounds(), self.enzmes(), self.reactions()], [])
    
    def _get_entries_by_type(self, type):
        return sorted(reduce(set.union, [entry.name.split() for entry in self.entries() if entry.type == type], set()))
    
    @cached_method
    def genes(self):
        """ Return all genes on the pathway
        """
        return self._get_entries_by_type("gene")
    
    @cached_method
    def compounds(self):
        """ Return all compounds on the pathway
        """
        return self._get_entries_by_type("compound")
    
    @cached_method
    def enzymes(self):
        """ Return all enzymes on the pathway
        """
        return self._get_entries_by_type("enzyme")
    
    @cached_method
    def orthologs(self):
        """ Return all orthologs on the pathway
        """
        return self._get_entries_by_type("ortholog")
    
    @cached_method
    def maps(self):
        """ Return all linked maps on the pathway
        """
        return self._get_entries_by_type("map")
    
    @cached_method
    def groups(self):
        """ Return all groups on the pathway
        """
        return self._get_entries_by_type("ortholog")
    
    def get_image(self):
        """ Return an image of the pathway
        """
        return self._get_image_filename()
    
    @classmethod
    def list(cls, organism):
        from . import api 
        kegg = api.CachedKeggApi()
        return kegg.list_pathways(organism)
    



"""
DBGET entry 
"""
from __future__ import absolute_import

__all__ = ["parser", "fields"]

from collections import defaultdict

from . import fields
from .parser import DBGETEntryParser

def entry_decorate(cls):
    """ Decorate the DBEntry subclass with properties for acessing
    the fields through the 'DBField._convert' interface
     
    """
    reserved_names_map = {"class": "class_", "def": "def_"}
    def construct_one(name):
        def get(self):
            field = getattr(self, name, None)
            if field is not None:
                return field._convert()
            else:
                return None
        return property(get, doc=name)
    
    def construct_multiple(name):
        def get(self):
            field = getattr(self, name, None)
            if field is not None:
                return [f._convert() for f in field]
            else:
                return None
        return property(get, doc=name)
    
    for name, field in cls.FIELDS:
        name_lower = name.lower()
        if not hasattr(cls, name_lower):
            if name in cls.MULTIPLE_FIELDS:
                prop = construct_multiple(name)
            else:
                prop = construct_one(name)
            setattr(cls, reserved_names_map.get(name_lower, name_lower), prop)
        
    return cls

class DBEntry(object):
    """ A DBGET entry object.
    """
    FIELDS = [("ENTRY", fields.DBEntryField),
              ]
    MULTIPLE_FIELDS = []
    
    def __init__(self, text=None):
        self._sections = {}
        if text is not None:
            self.parse(text)
            
    @property
    def entry_key(self):
        """ Primary entry key used for querying.
        """
        return self.entry.split(" ", 1)[0]
    
    def parse(self, text):
        parser = DBGETEntryParser()
        gen = parser.parse_string(text)
        entry_start = 0
        field_constructors = dict(self.FIELDS)
        sections = defaultdict(list)
        current = None
        current_subfield = None
        entry_fields = []
        for (event, title, text) in gen:
#            print event, title, text
            if event == DBGETEntryParser.SECTION_START:
                if title in field_constructors:
                    ftype = field_constructors[title]
                else:
                    ftype = fields.DBSimpleField
                current = ftype(text)
                if current.TITLE is None:
                    current.TITLE = title
            elif event == DBGETEntryParser.SECTION_END:
                entry_fields.append(current)
                current = None
            elif event == DBGETEntryParser.SUBSECTION_START:
                current_subfield = fields.DBSimpleField(text)
                current_subfield.TITLE = title
                if not isinstance(current, fields.DBFieldWithSubsections):
                    # Upgrade simple fields to FieldWithSubsection
                    new = fields.DBFieldWithSubsections(current.text)
                    new.TITLE = current.TITLE
                    current = new
                    
            elif event == DBGETEntryParser.SUBSECTION_END:
                current.subsections.append(current_subfield)
                current_subfield = None
            elif event == DBGETEntryParser.TEXT:
                if current_subfield is not None:
                    current_subfield.text += text
                elif current is not None:
                    current.text += text
            elif event == DBGETEntryParser.ENTRY_END:
                break
        
        self.fields = entry_fields
        self._consolidate()
        
    def _consolidate(self):
        """ Update mapping to field entries.
        """
        registered_fields = dict(self.FIELDS)
        multiple_fields = set(self.MULTIPLE_FIELDS)
        
        for field in self.fields:
            title = field.TITLE
            if title not in registered_fields:
                import warnings
                warnings.warn("Nonregisterd field %r in %r" % \
                                (title, type(self)))
            title_lower = title.lower()
            if title in multiple_fields:
                if not hasattr(self, title):
                    setattr(self, title, [])
                getattr(self, title).append(field)
            else:
                setattr(self, title, field)
        
    
    def __str__(self):
        return self.format()
    
    def format(self, section_indent=12):
        return "".join(f.format(section_indent)\
                       for f in self.fields)
        
    def get(self, key, default=None):
        raise NotImplementedError
    
        f = getattr(self, key, None)
        if f is not None:
            if key in self.MULTIPLE_FIELDS:
                return [f.text for f in f]
            else:
                return f.text
        else:
            return None



"""\
Protein-protein interactions
============================

This is python module for accessing PPI data.  
"""

from __future__ import absolute_import

import collections, gzip, posixpath, os, shutil, sqlite3, sys, urllib2, warnings, xml.dom.minidom as minidom
from collections import defaultdict
from operator import itemgetter

from Orange.orng import orngServerFiles
from Orange.utils import ConsoleProgressBar, lru_cache

from . import obiTaxonomy
from .obiKEGG import downloader
from .obiTaxonomy import pickled_cache

class PPIDatabase(object):
    """ A general interface for protein-protein interaction database access.
    
    An example useage::
        >>> ppidb = MySuperPPIDatabase()
        >>> ppidb.organisms() # List all organisms (taxids)
        ['...
        
        >>> ppidb.ids() # List all protein ids
        ['...
        
        >>> ppidb.ids(taxid="9606") # List all human protein ids.
        ['...
        
        >>> ppidb.links() # List all links 
        [('...
    """
    def __init__(self):
        pass
    
    def organisms(self):
        """ Return all organism taxids contained in this database.
        """
        raise NotImplementedError
    
    def ids(self, taxid=None):
        """ Return a list of all protein ids. If `taxid` is not None limit
        the results to ids from this organism only.
        
        """
        raise NotImplementedError
    
    def synonyms(self, id):
        """ Return a list of synonyms for primary `id`.
        """
        raise NotImplementedError
    
    def all_edges(self, taxid=None):
        """ Return a list of all edges. If taxid is not None return the
        edges for this organism only. 
        
        """
        raise NotImplementedError
    
    def edges(self, id1, id2=None):
        """ Return a list of all edges (a list of 3-tuples (id1, id2, score)).
        """
        raise NotImplementedError
    
    def all_edges_annotated(self, taxid=None):
        """ Return a list of all edges annotated. If taxid is not None
        return the edges for this organism only.
        
        """
        res = []
        for id in self.ids(taxid):
            res.extend(self.edges_annotated(id))
        return res
        
    def edges_annotated(self, id=None):
        """ Return a list of all edges annotated
        """
        raise NotImplementedError
     
    def search_id(self, name, taxid=None):
        """ Search the database for protein name. Return a list of matching 
        primary ids. Use `taxid` to limit the results to a single organism.
         
        """
        raise NotImplementedError
    
    def extract_network(self, ids):
        """
        """
        from Orange import network
        
        graph = network.Graph()
        for id in ids:
            graph.add_node(id, synonyms=",".join(self.synonyms(id)))
            
        for id in ids:
            edges = self.edges(id)
            
            for id1, id2, score in edges:
                graph.add_edge(id1, id2, weight=score)
                
        return graph
        
    
    @classmethod
    def download_data(self):
        """ Download the latest PPI data for local work.
        """
        raise NotImplementedError
    
class BioGRID(PPIDatabase):
    """ Access `BioGRID <http://thebiogrid.org>`_ PPI data.
    
    Example ::
    
        >>> biogrid = BioGRID()
        >>> print biogrid.organism() # Print a list of all organism ncbi taxis in BioGRID
        [u'10090',...
        
        >>> print biogrid.ids(taxid="9606") # Print a set of all human protein ids
        [u'110004'
        
        >>> print biogrid.synonyms("110004") # Print a list of all synonyms for protein id '110004' as reported by BioGRID
        [u'3803', u'CU464060.2', u'CD158b', u'p58.2', u'CD158B1', u'NKAT6']
        
        >>> 
        
    """
    
    SCHEMA = [("links", """\
        biogrid_interaction_id text, 
        biogrid_id_interactor_a text,
        biogrid_id_interactor_b text,
        experimental_system text,
        experimental_system_type text,
        author text,
        pubmed_id text,
        throughput text,
        score real,
        modification text,
        phenotypes text,
        qualifications text,
        tags text,
        source_database text
        """),
            ("proteins", """\
        biogrid_id_interactor text,
        entrez_gene_interactor text,
        systematic_name_interactor text,
        official_symbol_interactor text,
        synonyms_interactor text,
        organism_interactor text,
        """)]
    VERSION = "2.0"
    
    # All column names in the tab2 table. 
    FIELDS = ['biogrid_interaction_id',
              'entrez_gene_interactor_a',
              'entrez_gene_interactor_b',
              'biogrid_id_interactor_a',
              'biogrid_id_interactor_b',
              'systematic_name_interactor_a',
              'systematic_name_interactor_b',
              'official_symbol_interactor_a',
              'official_symbol_interactor_b',
              'synonyms_interactor_a',
              'synonyms_interactor_b',
              'experimental_system',
              'experimental_system_type',
              'author',
              'pubmed_id',
              'organism_interactor_a',
              'organism_interactor_b',
              'throughput',
              'score',
              'modification',
              'phenotypes',
              'qualifications',
              'tags',
              'source_database'
              ]
    
#    BioGRIDInteraction = collections.namedtuple("BioGRIDInteraction", " ".join(SCHEMA[0][1]))
#    BioGRIDInteractor = collections.namedtuple("BioGRIDInteractor", " ".join(SCHEMA[1][1]))
    
    DOMAIN = "PPI"
    SERVER_FILE = "BIOGRID-ALL.sqlite"
    
    def __init__(self):
        self.filename = orngServerFiles.localpath_download(self.DOMAIN, self.SERVER_FILE)
#        info = orngServerFiles.info(self.DOMAIN, self.SERVER_FILE)
        # assert version matches
        self.db = sqlite3.connect(self.filename)
        self.init_db_index()

    @lru_cache(1)
    def organisms(self):
        cur = self.db.execute("select distinct organism_interactor from proteins")
        return cur.fetchall()

    @lru_cache(3)
    def ids(self, taxid=None):
        """ Return a list of all protein ids (biogrid_id_interactors).
        If `taxid` is not None limit the results to ids from this organism
        only.
        
        """
        if taxid is None:
            cur = self.db.execute("""\
                select biogrid_id_interactor
                from proteins""")
        else:
            cur = self.db.execute("""\
                select biogrid_id_interactor
                from proteins
                where organism_interactor=?""", (taxid,))
        
        return [t[0] for t in cur.fetchall()]

    def synonyms(self, id):
        """ Return a list of synonyms for primary `id`.
        
        """
        cur = self.db.execute("""\
            select entrez_gene_interactor,
                   systematic_name_interactor,
                   official_symbol_interactor,
                   synonyms_interactor
            from proteins
            where biogrid_id_interactor=?
            """, (id,))
        rec = cur.fetchone()
        if rec:
            synonyms = list(rec[:-1]) + (rec[-1].split("|") if rec[-1] is not None else [])
            return [s for s in synonyms if s is not None]
        else:
            return []
    
    def all_edges(self, taxid=None):
        """ Return a list of all edges. If taxid is not None return the
        edges for this organism only. 
        
        """
        if taxid is not None:
            cur = self.db.execute("""\
                select biogrid_id_interactor_a, biogrid_id_interactor_a, score
                from links left join proteins on 
                    biogrid_id_interactor_a=biogrid_id_interactor or
                    biogrid_id_interactor_b=biogrid_id_interactor
                where organism_interactor=?
            """, (taxid,))
        else:
            cur = self.db.execute("""\
                select biogrid_id_interactor_a, biogrid_id_interactor_a, score
                from links
            """)
        edges = cur.fetchall()
        return edges
            
    def edges(self, id):
        """ Return a list of all interactions where id is a participant
        (a list of 3-tuples (id_a, id_b, score)).
        
        """
        
        cur = self.db.execute("""\
            select biogrid_id_interactor_a, biogrid_id_interactor_b, score
            from links
            where biogrid_id_interactor_a=? or biogrid_id_interactor_b=?
        """, (id, id))
        return cur.fetchall() 
        
    def all_edges_annotated(self, taxid=None):
        """ Return a list of all edges annotated. If taxid is not None
        return the edges for this organism only.
        
        """
        if taxid is not None:
            cur = self.db.execute("""\
                select *
                from links left join proteins on 
                    biogrid_id_interactor_a=biogrid_id_interactor or
                    biogrid_id_interactor_b=biogrid_id_interactor
                where organism_interactor=?
            """, (taxid,))
        else:
            cur = self.db.execute("""\
                select *
                from links
            """)
        edges = cur.fetchall()
        return edges
        
        
    def edges_annotated(self, id):
        """ Return a list of all links
        """
        cur = self.db.execute("""\
            select *
            from links
            where biogrid_id_interactor_a=? or biogrid_id_interactor_b=?
        """, (id, id))
        return cur.fetchall()
    
    def search_id(self, name, taxid=None):
        """ Search the database for protein name. Return a list of matching 
        primary ids. Use `taxid` to limit the results to a single organism.
         
        """
        # TODO: synonyms_interactor can contain multiple synonyms
        # (should create an indexed table of synonyms)
        if taxid is None:
            cur = self.db.execute("""\
                select biogrid_id_interactor
                from proteins
                where (biogrid_id_interactor=? or
                       entrez_gene_interactor=? or
                       systematic_name_interactor=? or
                       official_symbol_interactor=? or
                       synonyms_interactor=?)
            """, ((name,) * 5))
        else:
            cur = self.db.execute("""\
                select biogrid_id_interactor
                from proteins
                where (biogrid_id_interactor=? or
                       entrez_gene_interactor=? or
                       systematic_name_interactor=? or
                       official_symbol_interactor=? or
                       synonyms_interactor=?)
                      and organism_interactor=? 
            """, ((name,) * 5) + (taxid,))
        res = map(itemgetter(0), cur)
        return res 
    
    @classmethod
    def download_data(cls, address):
        """ Pass the address of the latest release (the tab2 format).
        """
        import urllib2, shutil, zipfile
        from StringIO import StringIO
        stream = urllib2.urlopen(address)
        stream = StringIO(stream.read())
        file = zipfile.ZipFile(stream)
        filename = file.namelist()[0]
        ppi_dir = orngServerFiles.localpath("PPI")
        file.extract(filename, ppi_dir)
        shutil.move(orngServerFiles.localpath("PPI", filename),
                    orngServerFiles.localpath("PPI", "BIOGRID-ALL.tab2"))
        filepath = orngServerFiles.localpath("PPI", "BIOGRID-ALL.tab2")
        cls.init_db(filepath)
        shutil.remove(filepath)
        
    @classmethod
    def init_db(cls, filepath):
        """ Initialize the sqlite data base from a BIOGRID-ALL.*tab2.txt file
        format.
        
        """
        dirname = os.path.dirname(filepath)
        lineiter = iter(open(filepath, "rb"))
        headers = lineiter.next() # read the first line
        
        con = sqlite3.connect(os.path.join(dirname, "BIOGRID-ALL.sqlite"))
        con.execute("drop table if exists links") # Drop old table
        con.execute("drop table if exists proteins") # Drop old table
        
        con.execute("""\
            create table links (
                biogrid_interaction_id text, 
                biogrid_id_interactor_a text,
                biogrid_id_interactor_b text,
                experimental_system text,
                experimental_system_type text,
                author text,
                pubmed_id text,
                throughput text,
                score real,
                modification text,
                phenotypes text,
                qualifications text,
                tags text,
                source_database text
            )""")
        
        con.execute("""\
            create table proteins (
                biogrid_id_interactor text,
                entrez_gene_interactor text,
                systematic_name_interactor text,
                official_symbol_interactor text,
                synonyms_interactor text,
                organism_interactor text
            )""")
        
        proteins = {}
        nulls = lambda values: [val if val != "-" else None for val in values]
        link_indices = [0, 3, 4, 11, 12, 13, 14, 17, 18, 19, 20, 21, 22, 23] # Values that go in the links table
        interactor_a_indices = [3, 1, 5, 7, 9, 15] # Values that go in the proteins table
        interactor_b_indices = [4, 2, 6, 8, 10, 16] # Values that go in the proteins table
        
        def processlinks(file):
            for line in file:
                if line != "\n":
                    fields = nulls(line.strip().split("\t"))
                    yield [fields[i] for i in link_indices]
                    interactor_a = [fields[i] for i in interactor_a_indices]
                    interactor_b = [fields[i] for i in interactor_b_indices]
                    proteins[interactor_a[0]] = interactor_a
                    proteins[interactor_b[0]] = interactor_b
        
        con.executemany("""\
            insert into links values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            """, processlinks(lineiter))
        
        con.executemany("""\
            insert into proteins values (?, ?, ?, ?, ?, ?)
            """, proteins.itervalues())
        con.commit()
        con.close()
        
    def init_db_index(self):
        """ Will create an indexes (if not already pressent) in the database
        for faster searching by primary ids.
        
        """
        self.db.execute("""\
        create index if not exists index_on_biogrid_id_interactor_a
           on links (biogrid_id_interactor_a)
        """)
        self.db.execute("""\
        create index if not exists index_on_biogrid_id_interactor_b
           on links (biogrid_id_interactor_b)
        """)
        self.db.execute("""\
        create index if not exists index_on_biogrid_id_interactor
           on proteins (biogrid_id_interactor)
        """)
        
from collections import namedtuple
from functools import partial

from .obiGeneMania import wget, copyfileobj

def chainiter(iterable):
    for sub_iter in iterable:
        for obj in sub_iter:
            yield obj

def chunks(iter, chunk_size=1000):
    chunk = []
    i = 0
    for obj in iter:
        chunk.append(obj)
        if len(chunk) == chunk_size:
            yield chunk
            chunk = []
    if chunk:
        # The remaining items if any
        yield chunk
    
STRINGInteraction = namedtuple("STRINGInteraciton",
        ["protein_id1", "protein_id2", "combined_score", "mode",
         "action", "score"])

class STRING(PPIDatabase):
    """ Access `STRING <http://www.string-db.org/>`_ PPI database.
    """
    
    DATABASE_SCHEMA = """\
    Database schema
    ---------------
    table `links`:
        - `protein_id1`: id (text)
        - `protein_id2`: id (text)
        - `score`: combined score (int)
        
    table `actions`:
        - `protein_id1`: id (text)
        - `protein_id2`: id (text)
        - `mode`: mode (text)
        - `action`: action type (text)
        - `score`: action score (int)
        
    table `proteins`:
        - `protein_id`: protein id in STRING (text) (in the form of {taxid}.{name})
        - `taxid`: organism taxid (text)
        
    table `aliases`:
        - `protein_id: id (text)
        - `alias`: protein alias (text)
        - `source`: protein alias source (text)
        
    """
    DOMAIN = "PPI"
    FILENAME = "string-protein.sqlite"
    VERSION = "2.0"
    
    # Mapping from obiTaxonomy.common_taxids() to taxids in STRING.
     
    TAXID_MAP = {"352472": "44689", # Dictyostelium discoideum
                 "562": None,
                 "2104": "272634", # Mycoplasma pneumoniae M129
                 "4530": "39947", # Oryza sativa Japonica Group
                 "4754": None,
                 "8355": None,
                 "4577": None
                 }
    
    
    def __init__(self):
        self.filename = orngServerFiles.localpath_download(self.DOMAIN, self.FILENAME)
        self.db = sqlite3.connect(self.filename)
        
    @classmethod
    def common_taxids(cls):
        common = obiTaxonomy.common_taxids()
        common = [cls.TAXID_MAP.get(id, id) for id in common]
        return filter(None, common)
    
    def organisms(self):
        """ Return all organism taxids contained in this database.
        """
        cur = self.db.execute("select distinct taxid from proteins")
        return [r[0] for r in cur.fetchall()]
    
    def ids(self, taxid=None):
        """ Return a list of all protein ids. If `taxid` is not None limit
        the results to ids from this organism only.
        
        """
        if taxid is not None:
            cur = self.db.execute("""\
                select protein_id
                from proteins
                where taxid=?
                """, (taxid,))
        else:
            cur = self.db.execute("""\
                select protein_id
                from proteins
                """)
        return [r[0] for r in cur.fetchall()]
    
    def synonyms(self, id):
        """ Return a list of synonyms for primary `id` as reported by STRING (proteins.aliases.{version}.txt file)
        """
        cur = self.db.execute("""\
            select alias
            from aliases
            where protein_id=?
            """, (id,))
        res = cur.fetchall()
        return [r[0] for r in res]
    
    def synonyms_with_source(self, id):
        """ Return a list of synonyms for primary `id` along 
        with its source as reported by STRING 
        (proteins.aliases.{version}.txt file)
        
        """
        cur = self.db.execute("""\
            select alias, source
            from aliases
            where protein_id=?
            """, (id,))
        res = cur.fetchall()
        return [(syn, set(source.split(" "))) \
                for syn, source in res]
    
    def all_edges(self, taxid=None):
        """ Return a list of all edges. If taxid is not None return the
        edges for this organism only.
        
        .. note:: This may take some time (and memory).
        
        """
        if taxid is not None:
            cur = self.db.execute("""\
                select links.protein_id1, links.protein_id2, score
                from links join proteins on
                    links.protein_id1=proteins.protein_id
                where taxid=?
                """, (taxid,))
        else:
            cur = self.db.execute("""\
                select protein_id1, protein_id1, score
                from links
                """)
        return cur.fetchall()
        
    def edges(self, id):
        """ Return a list of all edges (a list of 3-tuples (id1, id2, score)).
        """
        cur = self.db.execute("""\
            select protein_id1, protein_id2, score
            from links
            where protein_id1=?
            """, (id,))
        return cur.fetchall()
    
    def all_edges_annotated(self, taxid=None):
        """ Return a list of all edges annotated. If taxid is not None
        return the edges for this organism only.
        
        """
        res = []
        for id in self.ids(taxid):
            res.extend(self.edges_annotated(id))
        return res
        
    def edges_annotated(self, id):
        """ Return a list of all edges annotated.
        """
        cur = self.db.execute("""\
            select links.protein_id1, links.protein_id2, links.score, 
                   actions.action, actions.mode, actions.score
            from links left join actions on 
                   links.protein_id1=actions.protein_id1 and
                   links.protein_id2=actions.protein_id2
            where links.protein_id1=?
        """, (id,))
        return map(partial(apply,STRINGInteraction), cur.fetchall())
    
    def search_id(self, name, taxid=None):
        if taxid is None:
            cur = self.db.execute("""\
                select proteins.protein_id
                from proteins natural join aliases 
                where aliases.alias=?
            """, (name,))
        else:
            cur = self.db.execute("""\
                select proteins.protein_id
                from proteins natural join aliases 
                where aliases.alias=? and proteins.taxid=?
            """, (name, taxid))
        return map(itemgetter(0), cur)
        
    @classmethod
    def download_data(cls, version, taxids=None):
        """ Download the  PPI data for local work (this may take some time).
        Pass the version of the  STRING release e.g. v8.3.
        The resulting sqlite database will only contain the protein
        interactions for `taxids` (if None obiTaxonomy.common_taxids() will
        be used).
        
        """
        dir = orngServerFiles.localpath("PPI")

        base_url = "http://www.string-db.org/newstring_download/"
        links = base_url + "protein.links.{version}.txt.gz"
        actions = base_url + "protein.actions.{version}.txt.gz"
        aliases = base_url + "protein.aliases.{version}.txt.gz"
        
        wget(links.format(version=version), dir, progress=True)
        wget(actions.format(version=version), dir, progress=True)
        wget(aliases.format(version=version), dir, progress=True)
        
        links_filename = os.path.join(dir, "protein.links.{version}.txt".format(version=version))
        actions_filename = os.path.join(dir, "protein.actions.{version}.txt".format(version=version))
        aliases_filename = os.path.join(dir, "protein.aliases.{version}.txt".format(version=version))
        
        progress = ConsoleProgressBar("Extracting files:")
        progress(1.0)
        links_file = gzip.GzipFile(links_filename + ".gz", "rb")
        shutil.copyfileobj(links_file, open(links_filename, "wb"))
        
        progress(60.0)
        actions_file = gzip.GzipFile(actions_filename + ".gz", "rb")
        shutil.copyfileobj(actions_file, open(actions_filename, "wb"))
        
        progress(90.0)
        aliases_file = gzip.GzipFile(aliases_filename + ".gz", "rb")
        shutil.copyfileobj(aliases_file, open(aliases_filename, "wb"))
        progress.finish()
        
        cls.init_db(version, taxids)
        
    @classmethod
    def init_db(cls, version, taxids=None):
        """ Initialize the sqlite3 data base. `version` must contain a
        STRING release version e.g 'v8.3'. If `taxids` is not `None` it
        must contain a list of tax-ids in the STRING database for which
        to extract the interactions for.
        
        """
        def counter():
            i = 0
            while True:
                yield i
                i += 1
                
        protein_ids = defaultdict(counter().next)
        protein_taxid = {}

        dir = orngServerFiles.localpath(cls.DOMAIN)
        
        links_filename = os.path.join(dir, "protein.links.{version}.txt".format(version=version))
        actions_filename = os.path.join(dir, "protein.actions.{version}.txt".format(version=version))
        aliases_filename = os.path.join(dir, "protein.aliases.{version}.txt".format(version=version))
        
        links_file = open(links_filename, "rb")
        actions_file = open(actions_filename, "rb")
        aliases_file = open(aliases_filename, "rb")
        
        progress = ConsoleProgressBar("Processing links:")
        progress(0.0)
        filesize = os.stat(links_filename).st_size
        
        if taxids:
            taxids = set(taxids)
        else:
            taxids = set(cls.common_taxids())
                        
        con = sqlite3.connect(orngServerFiles.localpath(cls.DOMAIN, cls.FILENAME))
        
        with con:
            con.execute("drop table if exists links")
            con.execute("drop table if exists proteins")
            con.execute("drop table if exists actions")
            con.execute("drop table if exists aliases")
            
            con.execute("create table links (protein_id1 text, protein_id2 text, score int)")
            con.execute("create table proteins (protein_id text, taxid text)")
            con.execute("create table actions (protein_id1 text, protein_id2 text, mode text, action text, score int)")
            con.execute("create table aliases (protein_id text, alias text, source text)")
            
            header = links_file.readline() # read the header
            
            import csv
            reader = csv.reader(links_file, delimiter=" ")
            
            def read_links(reader, chunk_size=1000000):
                links = []
                i = 0
                split = str.split
                for p1, p2, score in reader:
                    if split(p1, ".", 1)[0] in taxids and split(p2, ".", 1)[0] in taxids:
                        links.append((intern(p1), intern(p2), int(score)))
                        if len(links) == chunk_size:
                            yield links
                            links = []
                    i += 1
                    if i % 1000 == 0: # Update the progress every 1000 lines
                        progress(100.0 * links_file.tell() / filesize)
                if links:
                    yield links
            
            for chunk in read_links(reader):
                con.executemany("insert into links values (?, ?, ?)", chunk)
            
            progress.finish()
            
            proteins = [res[0] for res in con.execute("select distinct protein_id1 from links")]
            progress = ConsoleProgressBar("Processing proteins:")
            
            def protein_taxids(proteins):
                protein_taxids = []
                for i, prot in enumerate(proteins):
                    taxid = prot.split(".", 1)[0]
                    protein_taxids.append((prot, taxid))
                    if i % 1000 == 0:
                        progress(100.0 * i / len(proteins))
                protein_taxids.sort()
                return protein_taxids
            
            con.executemany("insert into proteins values (?, ?)", protein_taxids(proteins))
            
            progress.finish()
            
            filesize = os.stat(actions_filename).st_size
            
            actions_file.readline() # read header
            
            progress = ConsoleProgressBar("Processing actions:")
            reader = csv.reader(actions_file, delimiter="\t")
            def read_actions(reader):
                actions = []
                i = 0
                split = str.split
                for p1, p2, mode, action, a_is_acting, score in reader:
                    if split(p1, ".", 1)[0] in taxids and split(p2, ".", 1)[0] in taxids:
                        actions.append((intern(p1), intern(p2), mode, action, int(score)))
                    i += 1
                    if i % 1000 == 0:
                        progress(100.0 * actions_file.tell() / filesize)
                actions.sort()
                return actions
            
            con.executemany("insert into actions values (?, ?, ?, ?, ?)", read_actions(reader))
            
            progress.finish()
            
            filesize = os.stat(aliases_filename).st_size
            aliases_file.readline() # read header
            
            progress = ConsoleProgressBar("Processing aliases:")
                            
            reader = csv.reader(aliases_file, delimiter="\t")
            def read_aliases(reader):
                i = 0
                for taxid, name, alias, source in reader:
                    if taxid in taxids:
                        yield (".".join([taxid, name]), 
                               alias.decode("utf-8", errors="ignore"), 
                               source.decode("utf-8", errors="ignore"),
                               )
                    i += 1
                    if i % 1000 == 0:
                        progress(100.0 * aliases_file.tell() / filesize)
                        
            con.executemany("insert into aliases values (?, ?, ?)", read_aliases(reader))
            
            progress.finish()

            print "Indexing the database"
            
            con.execute("""\
            create index if not exists index_link_protein_id1
                on links (protein_id1)""")
        
            con.execute("""\
                create index if not exists index_action_protein_id1
                    on actions (protein_id1)""")
            
            con.execute("""\
                create index if not exists index_proteins_id
                    on proteins (protein_id)""")
            
            con.execute("""\
                create index if not exists index_taxids
                    on proteins (taxid)""")
            
            con.execute("""\
                create index if not exists index_aliases_id
                    on aliases (protein_id)""")
            
            con.execute("""\
                create index if not exists index_aliases_alias
                    on aliases (alias)""")
            
            con.executescript("""
                DROP TABLE IF EXISTS version;
                
                CREATE TABLE version (
                     string_version text,
                     api_version text
                );""")
            
            con.execute("""
                INSERT INTO version
                VALUES (?, ?)""", (version, cls.VERSION))
            
        progress.finish()
        

STRINGDetailedInteraction = namedtuple("STRINGDetailedInteraction",
        ["protein_id1", "protein_id2", "combined_score", "mode",
         "action", "score", "neighborhood", "fusion", "cooccurence", 
         "coexpression", "experimental", "database", "textmining"
         ])

class STRINGDetailed(STRING):
    """  Access `STRING <http://www.string-db.org/>`_ PPI database.
    This class also allows access to subscores per channel.
    
    .. note:: This data is released under a 
        `Creative Commons Attribution-Noncommercial-Share Alike 3.0 License <http://creativecommons.org/licenses/by-nc-sa/3.0/>`_.
        If you want to use this data for commercial purpuses 
        you must get a license from STRING.
    
    """
    
    DATABASE_SCHEMA = """\
    DATABASE SCHEMA
    ===============
    
    table `evidence`:
        - `protein_id1`: protein id (text)
        - `protein_id2`: protein id (text)
        - `neighborhood`: score (int)
        - `fusion`: score (int)
        - `cooccurence`: score (int)
        - `coexpression`: score (int)
        - `experimental`: score (int)
        - `database`: score (int)
        - `textmining`: score (int)
        
    """
    FILENAME_DETAILED = "string-protein-detailed.sqlite"
    
    def __init__(self):
        STRING.__init__(self)
        db_file = orngServerFiles.localpath(self.DOMAIN, self.FILENAME)
        db_detailed_file = orngServerFiles.localpath_download(self.DOMAIN, self.FILENAME_DETAILED)
        self.db_detailed = sqlite3.connect(db_detailed_file)
        self.db_detailed.execute("ATTACH DATABASE ? as string", (db_file,))
        
    def edges_annotated(self, id):
        """ Return a list of all edges annotated.
        """
        edges = STRING.edges_annotated(self, id)
        edges_nc = []
        for edge in edges:
            id1, id2 = edge.protein_id1, edge.protein_id2
            cur = self.db_detailed.execute("""\
                SELECT neighborhood, fusion, cooccurence, coexpression,
                       experimental, database, textmining
                FROM evidence
                WHERE protein_id1=? AND protein_id2=?
                """, (id1, id2))
            res = cur.fetchone()
            if res:
                evidence = res
            else:
                evidence = [0] * 7
            edges_nc.append(
                STRINGDetailedInteraction(*(tuple(edge) + tuple(evidence)))
            )
        return edges_nc
        
    @classmethod
    def download_data(cls, version, taxids=None):
        import gzip, shutil, csv
        
        baseurl = "http://www.string-db.org/newstring_download/"
        links_filename = "protein.links.detailed.{version}.txt.gz".format(version=version)
        
        dir = orngServerFiles.localpath(cls.DOMAIN)
        local_filename = os.path.join(dir, links_filename)
         
        if not os.path.exists(local_filename):
            wget(baseurl + links_filename, dir, progress=True)
        else:
            print "Already downloaded - skiping"
            
        gz = gzip.open(os.path.join(dir, links_filename), "rb")

        # Strip .gz extension
        links_filename = os.path.join(dir, os.path.splitext(links_filename)[0])
        if not os.path.exists(links_filename):
            shutil.copyfileobj(gz, open(links_filename))
        
        cls.init_db(version, taxids)
            
    @classmethod
    def init_db(cls, version, taxids=None):
        import csv
        dir = orngServerFiles.localpath(cls.DOMAIN)
        
        links_filename = "protein.links.detailed.{version}.txt".format(version=version)
        links_filename = os.path.join(dir, links_filename)
        
        if taxids:
            taxids = set(taxids)
        else:
            taxids = set(cls.common_taxids())
        
        links_file = open(links_filename, "rb")
        
        con = sqlite3.connect(os.path.join(dir, cls.FILENAME_DETAILED))
        with con:
            con.execute("""\
                DROP TABLE IF EXISTS evidence
            """)
            
            con.execute("""\
                CREATE TABLE evidence(
                     protein_id1 TEXT,
                     protein_id2 TEXT,
                     neighborhood INTEGER, 
                     fusion INTEGER,
                     cooccurence INTEGER,
                     coexpression INTEGER,
                     experimental INTEGER,
                     database INTEGER,
                     textmining INTEGER
                    )
                """)
            
            links = csv.reader(links_file, delimiter=" ")
            links.next() # Read header
            filesize = os.stat(links_filename).st_size
            
            progress = ConsoleProgressBar("Processing links file:")
            progress(1.0)
            
            def read_links(reader, chunk_size=100000):
                split = str.split
                for links in chunks(reader, chunk_size):
                    chunk = []
                    for p1, p2, n, f, c, cx, ex, db, t, _ in links:
                        if split(p1, ".", 1)[0] in taxids and split(p2, ".", 1)[0] in taxids:
                            chunk.append((intern(p1), intern(p2), n, f, c, cx, ex, db, t))
                        
                    progress(100.0 * links_file.tell() / filesize)
                    if chunk:
                        yield chunk
            
            # The links are read in chunks for better performace 
            for chunk in read_links(links):
                con.executemany("""
                    INSERT INTO evidence 
                    VALUES  (?, ?, ?, ?, ?, ?, ?, ?, ?)
                    """, chunk)
        
            progress.finish()
            
            print "Indexing"
            con.execute("""\
                CREATE INDEX IF NOT EXISTS index_evidence
                    ON evidence (protein_id1, protein_id2)
            """)
            
            con.executescript("""
                DROP TABLE IF EXISTS version;
                
                CREATE TABLE version (
                     string_version text,
                     api_version text
                );
                """)
            
            con.execute("""
                INSERT INTO version
                VALUES (?, ?)""", (version, cls.VERSION))
        
        
class Interaction(object):
    def __init__(self, protein1, protein2, ref1=None, ref2=None, conf1=None, conf2=None):
        self.protein1, self.protein2 = protein1, protein2
        self.ref1, self.ref2 = ref1, ref2
        self.conf1, self.conf2 = conf1, conf2
        self.org1, self.org2 = None, None
    
class MIPS(object):
    VERSION = 1
    def __init__(self):
        self.load()
        
    def load(self):
        self.protein_names = defaultdict(set)
        self.refs = {}
        self.confidance = {}
        def process(element):
            d = {}
            participants = element.getElementsByTagName("proteinParticipant")
            proteins = []
            for protein in participants:
                interactor = protein.getElementsByTagName("proteinInteractor")[0]
                names = []
                for name in interactor.getElementsByTagName("shortLabel") + \
                            interactor.getElementsByTagName("fullName"):
                    names.append((name.tagName, name.childNodes[0].data))
                
                refs = []
                for ref in interactor.getElementsByTagName("primaryRef"):
                    refs += [(ref.tagName, ref.attributes.items())]
                org = dict(interactor.getElementsByTagName("organism")[0].attributes.items()).get("ncbiTaxId")
                conf = protein.getElementsByTagName("confidence")[0].attributes.items()
                proteins.append((names, refs, conf, org))
            interaction = Interaction(proteins[0][0][1][1], proteins[1][0][1][1])
            interaction.ref1, interaction.ref2 = proteins[0][1], proteins[1][1]
            interaction.conf1, interaction.conf2 = proteins[0][2], proteins[1][2]
            interaction.org1, interaction.org2 = proteins[0][3], proteins[1][3]
            
            self.protein_names[interaction.protein1].add(proteins[0][0][0][1])
            self.protein_names[interaction.protein2].add(proteins[1][0][0][1])
            
            return interaction 
            
        document = minidom.parse(orngServerFiles.localpath_download("PPI", "allppis.xml"))
        interactions = document.getElementsByTagName("interaction")
        self.interactions = [process(interaction) for interaction in interactions]
        
        self.protein_interactions = defaultdict(set)
        
        for inter in self.interactions:
            self.protein_names[inter.protein1] = dict(inter.ref1[0][1]).get("id")
            self.protein_names[inter.protein2] = dict(inter.ref2[0][1]).get("id")
            self.protein_interactions[inter.protein1].add(inter)
            self.protein_interactions[inter.protein2].add(inter) 
       
    def __iter__(self):
        return iter(self.interactions)
    
    @classmethod
    def download(cls):
        import urllib2, shutil
        src = urllib2.urlopen("http://mips.helmholtz-muenchen.de/proj/ppi/data/mppi.gz")
        dest = orngServerFiles.localpath("PPI", "mppi.gz")
        shutil.copyfileobj(src, open(dest, "wb"))
       
    @classmethod 
    @pickled_cache(None, [("PPI", "allppis.xml")], version=1)
    def _get_instance(cls):
        return MIPS()
    
    @classmethod
    def get_instance(cls):
        if not hasattr(cls, "_instance"):
            cls._instance= cls._get_instance()
        return cls._instance
    
def mips_interactions(protein = None):
    mips = MIPS.get_instance()
    if protein is None:
        return list(mips)
    else:
        return mips.protein_interactions.get(protein)

def mips_proteins():
    return set(MIPS.get_instance().protein_names.keys())

class BioGRIDInteraction(object):
    """ An object representing a BioGRID interaction. Each member of this object
    represents a data from a single column of BIOGRID-ALL.tab file.
    Attributes:
        - *interactor_a*    - BioGRID identifier
        - *interactor_b*    - BioGRID identifier
        - *official_symbol_a*    - An official symbol for *interactor_a*
        - *official_symbol_b*    - An official symbol for *interactor_b*
        - *aliases_for_a*    - Aliases separated by '|'
        - *aliases_for_b*    - Aliases separated by '|'
        - *experimental_system*     - Experimental system (see BioGRID documentation on www.thebiogrid.org for a list of valid entrys)
        - *source*    - 
        - *organism_a_id*    - NCBI Taxonomy identifier for *interactor_a*'s organism
        - *organism_b_id*    - NCBI Taxonomy identifier for *interactor_b*'s organism
    """
    __slots__ = ["interactor_a", "interactor_b", "official_symbol_a","official_symbol_b", "aliases_for_a", "aliases_for_b", "experimental_system", "source", "pubmed_id", "organism_a_id", "organism_b_id"]
    def __init__(self, line):
        for attr, val in zip(self.__slots__, line.split("\t")):
            setattr(self, attr, val)

class _BioGRID_Old(object):
    """ A BioGRID database interface
    Example::
        >>> ## finding all interactions for Homo sapiens sapiens
        >>> grid = BioGRID(case_insensitive=True)
        >>> proteins = proteins = biogrid.proteins() ## All proteins
        >>> proteins = [p for p in proteins if any(["9606" in [int.organism_a_id, int.organism_b_id] for int in grid.get(p)])]
    """
    VERSION = 1
    def __init__(self, case_insensitive=True):
#        warnings.warn("obiPPi._BioGRID_Old class is deprecated. Use obiPPI.BioGRID")
        self.case_insensitive = case_insensitive
        self._case = (lambda name: name.lower()) if self.case_insensitive else (lambda name: name)
        self.load()
        
    def load(self):
        text = open(orngServerFiles.localpath_download("PPI", "BIOGRID-ALL.tab"), "rb").read()
        text = text.split("SOURCE\tPUBMED_ID\tORGANISM_A_ID\tORGANISM_B_ID\n", 1)[-1]
        self.interactions = [BioGRIDInteraction(line) for line in text.split("\n") if line.strip()]
        
        self.protein_interactions = defaultdict(set)
        self.protein_names = {}
        
        case = self._case

        def update(keys, value, collection):
            for k in keys:
                collection.setdefault(k, set()).add(value)
                
        for inter in self.interactions:
            update(map(case, [inter.official_symbol_a] + inter.aliases_for_a.split("|")), case(inter.interactor_a), self.protein_names)
            update(map(case, [inter.official_symbol_b] + inter.aliases_for_b.split("|")), case(inter.interactor_b), self.protein_names)
            
            self.protein_interactions[case(inter.interactor_a)].add(inter)
            self.protein_interactions[case(inter.interactor_b)].add(inter)
            
        self.protein_interactions = dict(self.protein_interactions)

        if case("N/A") in self.protein_names:
            del self.protein_names[case("N/A")]
        
    def proteins(self):
        """ Return all protein names in BioGRID (from INTERACTOR_A, and INTERACTOR_B columns) 
        """
        return self.protein_interactions.keys()
            
    def __iter__(self):
        """ Iterate over all BioGRIDInteraction objects
        """
        return iter(self.interactions)
    
    def __getitem__(self, key):
        """ Return a list of protein interactions that a protein is a part of 
        """
        key = self._case(key)
#        keys = self.protein_alias_matcher.match(key)
        if key not in self.protein_interactions:
            keys = self.protein_names.get(key, [])
        else:
            keys = [key]
        if keys:
            return list(reduce(set.union, [self.protein_interactions.get(k, []) for k in keys], set()))
        else:
            raise KeyError(key)
    
    def get(self, key, default=None):
        """ Return a list of protein interactions that a protein is a part of
        """
        key = self._case(key)
#        keys = self.protein_alias_matcher.match(key)
        if key not in self.protein_interactions:
            keys = self.protein_names.get(keys, [])
        else:
            keys = [key] 
        if keys:
            return list(reduce(set.union, [self.protein_interactions.get(k, []) for k in keys], set()))
        else:
            return default
        
    @classmethod
    def get_instance(cls):
        if getattr(cls, "_instance", None) is None:
            cls._instance = _BioGRID_Old()
        return cls._instance
    
def biogrid_interactions(name=None):
    """Return a list of protein interactions (BioGRIDInteraction objects) that a protein is a part of
    """ 
    if name:
        return list(_BioGRID_Old.get_instance().get(name, set()))
    else:
        return _BioGRID_Old.get_instance().interactions
    
def biogrid_proteins():
    """ Return all protein names in BioGRID (from INTERACTOR_A, and INTERACTOR_B columns)
    """
    return _BioGRID_Old.get_instance().proteins()


if __name__ == "__main__":
    for protein in mips_proteins():
        print "Protein", protein, "interacts with", 
        print ",".join(set(reduce(list.__add__, [[inter.protein1, inter.protein2] for inter in mips_interactions(protein)], [])) -set([protein]))
            



"""
<name>PIPAx</name>
<description>Access data from PIPA RNA-Seq database.</description>
<icon>icons/PIPA.png</icon>
<priority>35</priority>
"""

from __future__ import absolute_import

import sys, os
from collections import defaultdict
import math
from datetime import date

from Orange.orng import orngEnviron
from Orange.OrangeWidgets import OWGUI
from Orange.OrangeWidgets.OWWidget import *

from .. import obiDicty

from .OWPIPA import (MyTreeWidgetItem, ListItemDelegate,
                    SelectionSetsWidget, SortedListWidget)

try:
    from ast import literal_eval
except ImportError:
    # Compatibility with Python 2.5
    literal_eval = eval


def tfloat(s):
    try:
        return float(s)
    except:
        return None


class MyTreeWidgetItem(QTreeWidgetItem):

    def __init__(self, parent, *args):
        QTreeWidgetItem.__init__(self, parent, *args)
        self.par = parent

    def __contains__(self, text):
        return any(text.upper() in str(self.text(i)).upper() \
                   for i in range(self.columnCount()))

    def __lt__(self, o1):
        col = self.par.sortColumn()
        if col in [8, 9]:  # WARNING: hardcoded column numbers
            return tfloat(self.text(col)) < tfloat(o1.text(col))
        else:
            return QTreeWidgetItem.__lt__(self, o1)


# set buffer file
bufferpath = os.path.join(orngEnviron.directoryNames["bufferDir"], "pipax")

try:
    os.makedirs(bufferpath)
except:
    pass

bufferfile = os.path.join(bufferpath, "database.sq3")


class SelectionByKey(object):
    """An object stores item selection by unique key values
    (works only for row selections in list and table models)
    Example::

        ## Save selection by unique tuple pairs (DisplayRole of column 1 and 2)
        selection = SelectionsByKey(itemView.selectionModel().selection(),
                                    key = (1,2))
        ## restore selection (Possibly omitting rows not present in the model)
        selection.select(itemView.selectionModel())

    """

    def __init__(self, itemSelection, name="", key=(0,)):
        self._key = key
        self.name = name
        self._selected_keys = []
        if itemSelection:
            self.setSelection(itemSelection)

    def _row_key(self, model, row):
        def key(row, col):
            return str(model.data(model.index(row, col),
                                  Qt.DisplayRole).toString())

        return tuple(key(row, col) for col in self._key)

    def setSelection(self, itemSelection):
        self._selected_keys = [self._row_key(ind.model(), ind.row()) \
                               for ind in itemSelection.indexes() \
                               if ind.column() == 0]

    def select(self, selectionModel):
        model = selectionModel.model()
        selectionModel.clear()
        for i in range(model.rowCount()):
            if self._row_key(model, i) in self._selected_keys:
                selectionModel.select(model.index(i, 0),
                    QItemSelectionModel.Select | QItemSelectionModel.Rows)

    def __len__(self):
        return len(self._selected_keys)


# Mapping from PIPAx.results_list annotation keys to Header names.
HEADER = [("_cached", ""),
          ("data_name", "Name"),
          ("species_name", "Species"),
          ("strain", "Strain"),
          ("Experiment", "Experiment"),
          ("genotype", "Genotype"),
          ("treatment", "Treatment"),
          ("growth", "Growth"),
          ("tp", "Timepoint"),
          ("replicate", "Replicate"),
          ("unique_id", "ID"),
          ("date_rnaseq", "Date RNAseq"),
          ("adapter_type", "Adapter"),
          ("experimenter", "Experimenter"),
          ("band", "Band"),
          ("polya", "Polya"),
          ("primer", "Primer"),
          ("shearing", "Shearing")
          ]

# Index of unique_id
ID_INDEX = 10

# Index of 'date_rnaseq'
DATE_INDEX = 11

SORTING_MODEL_LIST = \
    ["Strain", "Experiment", "Genotype",
     "Timepoint", "Growth", "Species",
     "ID", "Name", "Replicate"]


class OWPIPAx(OWWidget):
    settingsList = ["server", "excludeconstant", "username", "password",
                    "joinreplicates", "selectionSetsWidget.selections",
                    "columnsSortingWidget.sortingOrder", "currentSelection",
                    "log2", "experimentsHeaderState", "rtypei"]

    def __init__(self, parent=None, signalManager=None, name="PIPAx"):
        OWWidget.__init__(self, parent, signalManager, name)
        self.outputs = [("Example table", ExampleTable)]

        self.username = ""
        self.password = ""
        self.log2 = False
        self.rtypei = 0

        self.selectedExperiments = []
        self.buffer = obiDicty.BufferSQLite(bufferfile)

        self.searchString = ""
        self.excludeconstant = False
        self.joinreplicates = False
        self.currentSelection = None

        self.experimentsHeaderState = \
                dict(((name, False) for _, name in HEADER[:ID_INDEX + 1]))

        self.result_types = []
        self.mappings = {}

        self.controlArea.setMaximumWidth(250)
        self.controlArea.setMinimumWidth(250)

        OWGUI.button(self.controlArea, self, "Reload",
                     callback=self.Reload)
        OWGUI.button(self.controlArea, self, "Clear cache",
                     callback=self.clear_cache)

        b = OWGUI.widgetBox(self.controlArea, "Experiment Sets")
        self.selectionSetsWidget = SelectionSetsWidget(self)
        self.selectionSetsWidget.setSizePolicy(QSizePolicy.Preferred,
                                               QSizePolicy.Maximum)
        b.layout().addWidget(self.selectionSetsWidget)

        OWGUI.separator(self.controlArea)

        b = OWGUI.widgetBox(self.controlArea, "Sort output columns")
        self.columnsSortingWidget = SortedListWidget(self)
        self.columnsSortingWidget.setSizePolicy(QSizePolicy.Preferred,
                                                QSizePolicy.Maximum)
        b.layout().addWidget(self.columnsSortingWidget)
        sorting_model = QStringListModel(SORTING_MODEL_LIST)
        self.columnsSortingWidget.setModel(sorting_model)

        self.columnsSortingWidget.sortingOrder = \
                ["Strain", "Experiment", "Genotype", "Timepoint"]

        OWGUI.separator(self.controlArea)

        box = OWGUI.widgetBox(self.controlArea, 'Expression Type')
        self.expressionTypesCB = OWGUI.comboBox(box, self, "rtypei",
                items=[],
                callback=self.UpdateResultsList)

        OWGUI.checkBox(self.controlArea, self, "excludeconstant",
                       "Exclude labels with constant values"
                       )

        OWGUI.checkBox(self.controlArea, self, "joinreplicates",
                       "Average replicates (use median)"
                       )

        OWGUI.checkBox(self.controlArea, self, "log2",
                       "Logarithmic (base 2) transformation"
                       )

        self.commit_button = OWGUI.button(self.controlArea, self, "&Commit",
                                          callback=self.Commit)
        self.commit_button.setDisabled(True)

        OWGUI.rubber(self.controlArea)

        box = OWGUI.widgetBox(self.controlArea, "Authentication")

        OWGUI.lineEdit(box, self, "username", "Username:",
                       labelWidth=100,
                       orientation='horizontal',
                       callback=self.AuthChanged)

        self.passf = OWGUI.lineEdit(box, self, "password", "Password:",
                                    labelWidth=100,
                                    orientation='horizontal',
                                    callback=self.AuthChanged)

        self.passf.setEchoMode(QLineEdit.Password)

        OWGUI.lineEdit(self.mainArea, self, "searchString", "Search",
                       callbackOnType=True,
                       callback=self.SearchUpdate)

        self.headerLabels = [t[1] for t in HEADER]

        self.experimentsWidget = QTreeWidget()
        self.experimentsWidget.setHeaderLabels(self.headerLabels)
        self.experimentsWidget.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.experimentsWidget.setRootIsDecorated(False)
        self.experimentsWidget.setSortingEnabled(True)

        contextEventFilter = OWGUI.VisibleHeaderSectionContextEventFilter(
                            self.experimentsWidget, self.experimentsWidget
                            )

        self.experimentsWidget.header().installEventFilter(contextEventFilter)
        self.experimentsWidget.setItemDelegateForColumn(0,
                    OWGUI.IndicatorItemDelegate(self, role=Qt.DisplayRole)
                    )

        self.experimentsWidget.setAlternatingRowColors(True)

        self.connect(self.experimentsWidget.selectionModel(),
                 SIGNAL("selectionChanged(QItemSelection, QItemSelection)"),
                 self.onSelectionChanged)

        self.selectionSetsWidget.setSelectionModel(
                            self.experimentsWidget.selectionModel()
                            )

        self.mainArea.layout().addWidget(self.experimentsWidget)

        self.loadSettings()

        self.restoreHeaderState()

        self.connect(self.experimentsWidget.header(),
                     SIGNAL("geometriesChanged()"),
                     self.saveHeaderState)

        self.dbc = None

        self.AuthSet()

        QTimer.singleShot(100, self.UpdateExperiments)

        self.resize(800, 600)

    def AuthSet(self):
        if len(self.username):
            self.passf.setDisabled(False)
        else:
            self.passf.setDisabled(True)

    def AuthChanged(self):
        self.AuthSet()
        self.ConnectAndUpdate()

    def ConnectAndUpdate(self):
        self.Connect()
        self.UpdateExperiments(reload=True)

    def Connect(self):
        self.error(1)
        self.warning(1)

        def en(x):
            return x if len(x) else None

        self.dbc = obiDicty.PIPAx(buffer=self.buffer,
                                  username=en(self.username),
                                  password=self.password)

        #check password
        if en(self.username) != None:
            try:
                self.dbc.mappings(reload=True)
            except obiDicty.AuthenticationError:
                self.error(1, "Wrong username or password")
                self.dbc = None
            except Exception, ex:
                print "Error when contacting the PIPA database", ex
                import traceback
                print traceback.format_exc()
                try:  # maybe cached?
                    self.dbc.mappings()
                    self.warning(1, "Can not access database - using cached data.")
                except Exception, ex:
                    self.dbc = None
                    self.error(1, "Can not access database.")

    def Reload(self):
        self.UpdateExperiments(reload=True)

    def clear_cache(self):
        self.buffer.clear()
        self.Reload()

    def rtype(self):
        """Return selected result template type """
        if self.result_types:
            return self.result_types[self.rtypei][0]
        else:
            return "-1"

    def UpdateExperimentTypes(self):
        self.expressionTypesCB.clear()
        items = [desc for _, desc  in self.result_types]
        self.expressionTypesCB.addItems(items)
        self.rtypei = max(0, min(self.rtypei, len(self.result_types) - 1))

    def UpdateExperiments(self, reload=False):
        self.experimentsWidget.clear()
        self.items = []

        self.progressBarInit()

        if not self.dbc:
            self.Connect()

        mappings = {}
        result_types = []
        sucind = False  # success indicator for database index

        try:
            mappings = self.dbc.mappings(reload=reload)
            result_types = self.dbc.result_types(reload=reload)
            sucind = True
        except Exception, ex:
            try:
                mappings = self.dbc.mappings()
                result_types = self.dbc.result_types()
                self.warning(0, "Can not access database - using cached data.")
                sucind = True
            except Exception, ex:
                self.error(0, "Can not access database.")

        if sucind:
            self.warning(0)
            self.error(0)

        self.mappings = mappings
        self.result_types = result_types

        self.UpdateExperimentTypes()

        results_list = {}
        try:
            results_list = self.dbc.results_list(self.rtype(), reload=reload)
        except Exception, ex:
            try:
                results_list = self.dbc.results_list(self.rtype())
            except Exception, ex:
                self.error(0, "Can not access database.")

        self.results_list = results_list
        mappings_key_dict = dict(((m["data_id"], m["id"]), key) \
                                 for key, m in mappings.items())

        def mapping_unique_id(annot):
            """Map annotations dict from results_list to unique
            `mappings` ids.
            """
            data_id, mappings_id = annot["data_id"], annot["mappings_id"]
            return mappings_key_dict[data_id, mappings_id]

        elements = []
        pos = 0

        for r_id, r_annot in self.results_list.items():
            pos += 1
            d = defaultdict(lambda: "?", r_annot)
            row_items = [""] + [d.get(key, "?") for key, _ in HEADER[1:]]
            date_string = row_items[DATE_INDEX]
            try:
                time_dict = literal_eval(date_string)
            except Exception:
                time_dict = {}

            if time_dict and "dateUTC" in time_dict and \
                    "monthUTC" in time_dict and "fullYearUTC" in time_dict:
                date_rna = date(time_dict["fullYearUTC"],
                                time_dict["monthUTC"] + 1,  # Why is month 0 based?
                                time_dict["dateUTC"])

                row_items[DATE_INDEX] = date_rna.strftime("%x")

            row_items[ID_INDEX] = mapping_unique_id(r_annot)
            elements.append(row_items)

            ci = MyTreeWidgetItem(self.experimentsWidget, row_items)

            self.items.append(ci)

        for i in range(len(self.headerLabels)):
            self.experimentsWidget.resizeColumnToContents(i)

        # which is the ok buffer version
        # FIXME: what attribute to use for version?
        self.wantbufver = \
            lambda x, ad=self.results_list: \
                defaultdict(lambda: "?", ad[x])["date"]

        self.wantbufver = lambda x: "0"

        self.UpdateCached()

        self.progressBarFinished()

        if self.currentSelection:
            self.currentSelection.select(self.experimentsWidget.selectionModel())

        self.handle_commit_button()

    def UpdateResultsList(self, reload=False):
        results_list = {}
        try:
            results_list = self.dbc.results_list(self.rtype(), reload=reload)
        except Exception:
            results_list = self.dbc.results_list(self.rtype())
        self.results_list = results_list
        self.UpdateCached()

    def UpdateCached(self):
        if self.wantbufver and self.dbc:
            fn = self.dbc.download_key_function()
            result_id_key = dict(((m["data_id"], m["mappings_id"]), key) \
                                 for key, m in self.results_list.items())

            for item in self.items:
                c = str(item.text(10))
                mapping = self.mappings[c]
                data_id, mappings_id = mapping["data_id"], mapping["id"]
                r_id = result_id_key[data_id, mappings_id]
                # Get the buffered version
                buffered = self.dbc.inBuffer(fn(r_id))
                value = " " if buffered == self.wantbufver(r_id) else ""
                item.setData(0, Qt.DisplayRole, QVariant(value))

    def SearchUpdate(self, string=""):
        for item in self.items:
            item.setHidden(not all(s in item \
                                   for s in self.searchString.split())
                           )

    def Commit(self):
        if not self.dbc:
            self.Connect()

        pb = OWGUI.ProgressBar(self, iterations=100)

        table = None

        ids = []
        for item in self.experimentsWidget.selectedItems():
            unique_id = str(item.text(10))
            annots = self.mappings[unique_id]
            ids.append((annots["data_id"], annots["id"]))

        transfn = None
        if self.log2:
            transfn = lambda x: math.log(x + 1.0, 2)

        reverse_header_dict = dict((name, key) for key, name in HEADER)

        hview = self.experimentsWidget.header()
        shownHeaders = [label for i, label in \
                        list(enumerate(self.headerLabels))[1:] \
                        if not hview.isSectionHidden(i)
                        ]

        allowed_labels = [reverse_header_dict.get(label, label) \
                          for label in shownHeaders]

        if self.joinreplicates and "id" not in allowed_labels:
            # need 'id' labels in join_replicates for attribute names
            allowed_labels.append("id")

        if len(ids):
            table = self.dbc.get_data(ids=ids, result_type=self.rtype(),
                          callback=pb.advance,
                          exclude_constant_labels=self.excludeconstant,
#                          bufver=self.wantbufver,
                          transform=transfn,
                          allowed_labels=allowed_labels)

            if self.joinreplicates:
                table = obiDicty.join_replicates(table,
                    ignorenames=["replicate", "data_id", "mappings_id",
                                 "data_name", "id", "unique_id"],
                    namefn=None,
                    avg=obiDicty.median
                    )

            # Sort attributes
            sortOrder = self.columnsSortingWidget.sortingOrder

            def sorting_key(attr):
                atts = attr.attributes
                return tuple([atts.get(reverse_header_dict[name], "") \
                              for name in sortOrder])

            attributes = sorted(table.domain.attributes,
                                key=sorting_key)

            domain = orange.Domain(attributes, table.domain.classVar)
            domain.addmetas(table.domain.getmetas())
            table = orange.ExampleTable(domain, table)

            from Orange.orng.orngDataCaching import data_hints
            data_hints.set_hint(table, "taxid", "352472")
            data_hints.set_hint(table, "genesinrows", False)

            self.send("Example table", table)

            self.UpdateCached()

        pb.finish()

    def onSelectionChanged(self, selected, deselected):
        self.handle_commit_button()

    def handle_commit_button(self):
        self.currentSelection = \
            SelectionByKey(self.experimentsWidget.selectionModel().selection(),
                           key=(1, 2, 3, 10))
        self.commit_button.setDisabled(not len(self.currentSelection))

    def saveHeaderState(self):
        hview = self.experimentsWidget.header()
        for i, label in enumerate(self.headerLabels):
            self.experimentsHeaderState[label] = hview.isSectionHidden(i)

    def restoreHeaderState(self):
        hview = self.experimentsWidget.header()
        state = self.experimentsHeaderState
        for i, label in enumerate(self.headerLabels):
            hview.setSectionHidden(i, state.get(label, True))
            self.experimentsWidget.resizeColumnToContents(i)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    obiDicty.verbose = True
    w = OWPIPAx()
    w.show()
    app.exec_()
    w.saveSettings()
